# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" 
Package aiming at making common operations "easy".
"""

import logging
logger = logging.getLogger(__name__)

from collections import namedtuple, OrderedDict, Counter, defaultdict
import collections
from importlib import import_module
from railroadtracks import core, unifex, rnaseq, hortator
import os, shutil, time
import json
import tempfile
import itertools, operator

from railroadtracks.easy.tasks import Task, TaskSet, DbID
from railroadtracks.easy.tasks import (_TASK_DONE,
                                       _TASK_TODO,
                                       _TASK_INPROGRESS,
                                       _TASK_FAILED)

from railroadtracks.easy.execution import (DummyExecution,
                                           IterativeExecution,
                                           MultiprocessingExecution)

TASK_DONE = _TASK_DONE
TASK_TODO = _TASK_TODO
TASK_INPROGRESS = _TASK_INPROGRESS
TASK_FAILED = _TASK_FAILED

class Asset(object):
    """ An asset is either used by a task (then it is a source asset) or is produced
    by a task (then it is a target asset). """
    __slots__ = ('_asset_id', '_entity', '_project')
    def __init__(self, project, savedentity, asset_id):
        self._asset_id = asset_id
        self._entity = savedentity
        self._project = project

    @property
    def id(self):
        return self._asset_id

    @property
    def entity(self):
        """ Instance of :class:`SavedEntityAbstract` 
        (or rather a child class thereof). """
        return self._entity
    @property
    def project(self):
        return self._project

    @property
    def parenttask(self):
        return self.project.persistent_graph.get_parenttask_of_storedentity(self.id)
        

ActivityCount = namedtuple('ActivityCount', 'count name status')
class ActivityDoneCount(object):
    __slots__ = ['count_done', 'count_failed', 'total', 'name']
    def __init__(self, name, count_done=0, count_failed=0, total=0):
        self.count_done = count_done
        self.count_failed = count_failed
        self.total = total
        self.name = name


def dict_project_view_storage(project):
    if project.newproject:
        status = 'New "railroadtracks" project'
    else:
        status = 'Reopened existing "railroadtracks" project'
    statvfs= os.statvfs(project.wd)

    avail = statvfs.f_bavail * statvfs.f_frsize / (1024 ** 3)
    total = statvfs.f_blocks * statvfs.f_frsize / (1024 ** 3)
    used = (statvfs.f_blocks - statvfs.f_bfree) * statvfs.f_frsize / (1024 ** 3)

    taskstatuscount = project.persistent_graph.nconcrete_steps_status
    totaltasks = sum(x.count for x in taskstatuscount)

    d = {'status': status, 
         'wd': project.wd, 
         'wdspace_avail': avail,
         'wdspace_used': used,
         'wdspace_total': total,
         'db_fn': project.db_fn,
         'db_fn_size': '{:,}'.format(os.stat(project.db_fn).st_size / (1024 ** 2)),
         'taskstatuscount': taskstatuscount,
         'totaltasks': totaltasks}
    return d

def dict_project_view_activities(project, done=hortator._TASK_DONE, failed=hortator._TASK_FAILED):
    cursor = project.persistent_graph.connection.cursor()
    sql = """
SELECT count(step_activity_id), step_activity.label, ss.label
FROM step_activity
LEFT JOIN step_type2activity AS sta ON sta.step_activity_id=step_activity.id
LEFT JOIN step_type ON sta.step_type_id=step_type.id
LEFT JOIN step_variant AS sv ON sv.step_type_id=step_type.id
LEFT JOIN step_concrete AS sc ON sc.step_variant_id=sv.id
LEFT JOIN step_status AS ss ON ss.id=sc.step_status_id
GROUP BY step_activity_id, step_status_id
ORDER BY step_activity_id
    """
    cursor.execute(sql)
    counts = tuple(ActivityCount(*x) for x in cursor.fetchall())
    activities = list()
    for name, grp in itertools.groupby(counts, operator.attrgetter('name')):
        adc = ActivityDoneCount(name)
        for ac in grp:
            if ac.status == done:
                adc.count_done = ac.count
            elif ac.status == failed:
                adc.count_failed = ac.count
            adc.total += ac.count
        activities.append(adc)
    d = {'activities': activities }
    return d



def dict_project_view_results(project):
    results = project.persistent_graph.iter_finaltargets()
    ct = dict()
    for result in results:
        if result.status_label == _TASK_DONE:
            count_done = 1
        else:
            count_done = 0
        adc = ct.get(result.stored_entity_classname)
        if adc is None:
            adc = ActivityDoneCount(result.stored_entity_classname, 
                                    count_done=count_done,
                                    total = 1)
            ct[result.stored_entity_classname] = adc
        else:
            adc.count_done += count_done
            adc.total += 1
    res = {'results': ct.values()}
    return res

def dict_assetset_view(assetset):
    l = len(assetset)
    AssetSetElement = namedtuple('AssetSetElement', 'isdefined name cls')
    d = {'super_repr': super(core.AssetSet, assetset).__repr__(),
         'elements': []}
    for i in range(l):
        isdefined = getattr(assetset, 
                            assetset._sources[i].name)._defined
        d['elements'].append(AssetSetElement(isdefined,
                                             assetset._sources[i].name,
                                             str(assetset._sources[i].cls)))
    return d

def str_progressbar(val, maxval, width=20, fillchar="#",
                    emptychar=' '):
    assert val <= maxval
    nfill = int(round(1.0 * val / maxval * width))
    res = fillchar * nfill
    if emptychar is not None:
        res += emptychar * (width-nfill)
    return res

def str_project_view_storage(project, width=10):
    d = dict_project_view_storage(project)
    for k in ('wdspace_avail', 'wdspace_total'):
        d[k] = '{:,.2f}'.format(d[k])
    res = """Status: %(status)s
Working directory: %(wd)s
Storage
  Available: %(wdspace_avail)s GB
  Total:     %(wdspace_total)s GB
Tasks
  Total:     %(totaltasks)i
""" % d
    return res


def str_project_view_activities(project, 
                                done=hortator._TASK_DONE, failed=hortator._TASK_FAILED,
                                width=20):
    d = dict_project_view_activities(project, done=done, failed=failed)
    if len(d['activities']) == 0:
        name_len = 0
    else:
        name_len = max(len(a.name) for a in d['activities'])
    name_len = max(name_len, len('Activity')) # column header
    res = list(('%s    %s' %('Activity'.center(name_len),'Progress'),))
    for a in d['activities']:
        res.append('%s    |%s| %.2f%% (%s/%s - failed: %s)' % (a.name.ljust(name_len), 
                                                  str_progressbar(a.count_done, a.total, width=width),
                                                  100.0*a.count_done/a.total,
                                                  '{:,}'.format(a.count_done),
                                                  '{:,}'.format(a.total),
                                                  '{:,}'.format(a.count_failed))
               )
    res.append('')
    return os.linesep.join(res)

def str_project_view_results(project, width=20):
    d = dict_project_view_results(project)
    if len(d['results']) == 0:
        name_len = 0
    else:
        name_len = max(len(r.name) for r in d['results'])
    name_len = max(name_len, len('Result Type')) # column header
    res = list(('%s    %s' %('Result Type'.center(name_len),'Progress'),))
    for r in d['results']:
        res.append('%s    |%s| %.2f%%' % (r.name.ljust(name_len), 
                                          str_progressbar(r.count_done, r.total, width=width),
                                          100.0*r.count_done/r.total))
    res.append('')
    return os.linesep.join(res)
    return res

def str_project_view(project):
    res = tuple((x(project) for x in (str_project_view_storage,
                                      str_project_view_activities,
                                      str_project_view_results)))
    return ('--%s' % os.linesep).join(res)


class Project(object):
    """ A project, that is a directory containing data as well as a persistent storage of steps and how derived data and final results were obtained. """

    def __init__(self, model, wd='railroadtracks_project', db_fn=None, force_create = False, 
                 isolation_level=None, cached=False):
        """
        :param model: module with the definition of the `model` (that is the classes defining the type of tasks computed)
        :param wd: Name of a working directory (where all intermediate and final results will be saved.
        :type wd: :class:`str`
        :param db_fn: Name of a file name for the database. If None, use the file "railroadtracks.db" in the directory specified in "wd".
        :type db_fn: :class:`str` or None
        :param isolation_level: passed to the constructor for :class:`PersistentTaskGraph`
        :rtype: a tuple with (:class:`hortator.StepGraph`, working directory as a :class:`str`, file name for the database and a :class:`str`)
        """

        # check the model
        if not hasattr(model, 'ACTIVITY'):
            raise ValueError('The parameter `model` does not appear to be a model (ACTIVITY missing).')

        newproject = force_create
        # derived data files and final results will be in a temporary directory
        if not os.path.isdir(wd):
            raise ValueError("The working directory '%s' should be a directory" % wd)

        # create the database
        if db_fn is None:
            db_fn = os.path.join(wd, 'railroadtracks.db')

        if not os.path.exists(db_fn):
            newproject = True
        if cached:
            cls_persistent = hortator.CachedPersistentTaskGraph
        else:
            cls_persistent = hortator.PersistentTaskGraph

        persistent_graph = cls_persistent(db_fn, model, wd=wd, force_create=force_create,
                                          isolation_level=isolation_level)

        self._persistent_graph = persistent_graph
        # cache the dependency graph
        cache = hortator.StepGraph(persistent_graph)
        self._cache = cache
        self._wd = wd
        self._db_fn = db_fn
        self._newproject = newproject

    @property
    def persistent_graph(self):
        """ Access to the persistent graph. This is will be slower than access the cached version, but will
        always be accurate."""
        return self._persistent_graph

    @property
    def cache(self):
        """ Cache for the dependency graph. This may not reflect changes made by an independent access
        to the project or a direct access to the persistent graph (see :attr:`persistent_graph`) but
        will be much faster and the preferred way to access the graph when there is only one concurrent access.
        """
        return self._cache

    @property
    def wd(self):
        """Working directory. The directory in which derived data is stored."""
        return self._wd

    @property
    def db_fn(self):
        """Path to the database file"""
        return self._db_fn

    @property
    def newproject(self):
        """Tell whether this is a new project, rather than a existing project (re)opened. """
        return self._newproject

    @property
    def model(self):
        """Model used in the project."""
        return self.persistent_graph._model

    def __repr__(self):
        res = """
%s
Working directory: %s
Database file: %s
Number of recorded steps: %i
"""
        if self.newproject:
            status = 'New project'
        else:
            status = 'Reopened existing project'
        return res % (status, self.wd, self.db_fn,
                      self.persistent_graph.nconcrete_steps)

    def __str__(self):
        return str_project_view(self)


    def get_targetsofactivity(self, activity):
        """ 
        Retrieve the targets of steps performing a specific activity.
        (calls the method of the same name in the contained :class:`PersistentTaskGraph`)
        :param activity: an activity
        :type activity: :class:`Enum`
        """
        return self.persistent_graph.get_targetsofactivity(activity)

    def get_targetsoftype(self, obj):
        """ 
        Retrieve the targets having a given type.
        (calls the method of the same name in the contained :class:`PersistentTaskGraph`)
        :param obj: a type, an instance, or a type name
        :type obj: :class:`type`, :class:`Object`, or :class:`str`
        """
        if isinstance(obj, str):
            # name of a class
            cls = getattr(self.persistent_graph._model, obj)
        elif isinstance(obj, core.SavedEntityAbstract):
            cls = type(obj)
        elif issubclass(obj, core.SavedEntityAbstract):
            cls = obj
        else:
            raise ValueError("Invalid parameter.")
        return self.persistent_graph.get_targetsoftype(cls.__name__)

    def get_tasksoftype(self, taskorstep):
        """
        Retrieve tasks matching the type of "taskorstep".

        The parameter 'taskorstep' can be either:
        - a class inheriting from StepAbstract
        - an instance of Task
        - an instance of StepAbstract
        In the two last cases, the path to the executable and the version number
        are used to filter the tasks returned.
        """
        cache = self.persistent_graph
        cursor = cache.connection.cursor()

        stepvariant_ids = None
        if issubclass(taskorstep, core.StepAbstract):
            cls = taskorstep
            step_type_id = cache.id_step_type(cls.activities)
            sql = """
            SELECT id
            FROM step_variant
            WHERE step_type_id=?
            AND cls=?
            """
            #FIXME: building the 'cls' string the SQL string
            #       should be factored out as a function
            #       in hortator.py
            cursor.execute(sql, (step_type_id.id, '.'.join((cls.__module__,
                                                         cls.__name__))))
            stepvariant_ids = tuple(x[0] for x in cursor.fetchall())
        else:
            if isinstance(taskorstep, Task):
                step = taskorstep.call.step
            elif isinstance(taskorstep, core.StepAbstract):
                step = taskorstep
            stepvariant_ids = (cache.id_step_variant(step, step.activities).id, )

        if stepvariant_ids is None:
            raise ValueError('"taskorstep" can be either a Task or a core.StepAbstract')

        taskset = TaskSet()

        sql = """
        SELECT step_concrete.id
        FROM step_concrete
        WHERE step_concrete.step_variant_id=?
        """

        for sv_id in stepvariant_ids:
            cursor.execute(sql, (sv_id,))
            for res in cursor.fetchall():
                task_id = res[0]
                taskset.add(self.get_task(task_id))

        return taskset

    def get_taskswithactivity(self, activity):
        """
        Retrieve tasks matching the given activity.
        """
        cache = self.persistent_graph
        cursor = cache.connection.cursor()

        taskset = TaskSet()

        sql = """
        SELECT step_concrete.id
        FROM step_concrete
        INNER JOIN step_variant
        ON step_variant.id=step_concrete.step_variant_id
        INNER JOIN step_type
        ON step_type.id=step_variant.step_type_id
        INNER JOIN step_type2activity
        ON step_type2activity.step_type_id=step_type.id
        INNER JOIN step_activity
        ON step_activity.id=step_type2activity.step_activity_id
        WHERE step_activity.label=?
        """

        cursor.execute(sql, (activity.value,))
        for res in cursor.fetchall():
            task_id = res[0]
            taskset.add(self.get_task(task_id))

        return taskset

    def iter_srcassets(self, task):
        """ Return the source files for a given task
        (calls the method of the same name in the contained :class:`PersistentTaskGraph`)
        :param task: a :class:`Task`.
        """
        for x in self.persistent_graph.get_srcassets(task.task_id):
            yield Asset(self, x.resurrect(self.model), DbID(x.id))
            

    def iter_targetassets(self, task):
        """ Return the target files for a given task
        (calls the method of the same name in the contained :class:`PersistentTaskGraph`)
        :param task: a :class:`Task`.
        """
        for x in self.persistent_graph.get_targetassets(task.task_id):
            yield Asset(self, x.resurrect(self.model), DbID(x.id))

    def add_task(self, step, assets, parameters=(), tag = 1):
        """
        Add a task to the project. If any of the assets' targets is not defined,
        it will be defined automatically.

        :param step:
        :type param: :class:`StepAbstract`
        :param assets:
        :param parameters:
        :param tag: a tag (to differentiate repetitions of the exact same task)
        """
        step_concrete_id = self.cache.add(step, assets, parameters=tuple(parameters), tag=tag)
        call = unifex.Call(step, assets, parameters)
        task = Task(self, call, step_concrete_id)
        return task


    def get_task(self, task_id):
        """
        Given a task ID, retrieve the associated task.
        """
        
        if not hasattr(task_id, 'id'):
            task_id = DbID(task_id)

        task_dbentry = self.persistent_graph._get_stepconcrete(task_id)
        # get the step
        if '.' in task_dbentry.clsname:
            module_name, cls_name = task_dbentry.clsname.rsplit('.', 1)
        else:
            module_name = self.persistent_graph._model
            cls_name = task_dbentry.clsname
        module = import_module(module_name)
        cls = getattr(module, cls_name)
        step = cls(task_dbentry.executable)
        # get the assets
        sources = [None, ]*len(cls.Assets.Source._fields)
        model = self.persistent_graph._model
        for i, x in enumerate(self.persistent_graph.get_srcassets(task_id)):
            src_i = cls.Assets.Source._fields.index(x.label)
            sources[src_i] = x.resurrect(model)
        targets = [None, ]*len(cls.Assets.Target._fields)
        for i, x in enumerate(self._persistent_graph.get_targetassets(task_id)):
            tgt_i = cls.Assets.Target._fields.index(x.label)
            targets[tgt_i] = x.resurrect(model)
        assets = cls.Assets(cls.Assets.Source(*sources),
                            cls.Assets.Target(*targets))
        # get the parameters
        parameters = json.loads(task_dbentry.parameters)
        # build the call
        call = unifex.Call(step, assets, parameters)
        # build the task
        task = Task(self, call, task_dbentry.id)
        return task

    def add_asset(self, asset):
        if hasattr(asset, 'iter_storedentities'):
            raise NotImplementedError('Not yet implemented')
            self.persistent_graph.id_stored_sequence()
        else:
            seid = self.persistent_graph.id_stored_entity(type(asset), asset.name)
            return hortator.StoredEntity(seid.id, None, asset.__class__.__name__, asset.name)

def call_factory(project, step_concrete_id, stepobj, assets, parameters=()):
    """ 
    :param stepobj: Step object
    :type stepobj: instance of :class:`StepAbstract`
    :param assets: assets
    :type assets: instance of :class:`AssetsStep`
    :rtype: :class:`function`
    """
    def call():
        import hortator
        cmd, returncode = stepobj.run(assets, parameters=parameters)
        if returncode != 0:
            raise RuntimeError('Error (return code %i) while running: %s' % (returncode, ' '.join(cmd)))
        project.persistent_graph.step_concrete_state(step_concrete_id,
                                                hortator._TASK_STATUS_LIST[hortator._TASK_DONE])
    return call




def _delete_tasks_in_set(taskset):
    """ Delete a set of tasks. """
    # get the target assets associated with the task
    #db = task.project.persistent_graph
    #storedentitities = list()
    #for task in taskset:
    #    storedentities.extend(db.get_targetassets(task.id))

    sql_create_task_tmp = """
    CREATE TEMP TABLE task_to_delete (
      id INTEGER PRIMARY KEY
    )"""

    sql_add_task_id_to_delete = """
    INSERT INTO task_to_delete
    (id)
    VALUES
    (?)
    """

    sql_create_stored_entity_tmp = """
    CREATE TEMP TABLE se_to_delete
    AS
      SELECT stored_entity.id AS se_id
      FROM stored_entity 
      INNER JOIN step_concrete2targetfile
      ON stored_entity.id=step_concrete2targetfile.stored_entity_id
      INNER JOIN task_to_delete
      ON task_to_delete.id=step_concrete2targetfile.step_concrete_id
      WHERE se_id IN (SELECT task_to_delete.id FROM task_to_delete)"""

    sql_delete_src_associations = """
    DELETE FROM 
    step_concrete2targetfile
    WHERE step_concrete_id IN (SELECT id from task_to_delete)
    """

    sql_delete_target_association = """
    DELETE FROM 
    step_concrete2targetfile
    WHERE step_concrete_id IN (SELECT id from task_to_delete)
    """

    sql_delete_src_association = """
    DELETE FROM 
    step_concrete2srcfile
    WHERE step_concrete_id IN (SELECT id from task_to_delete)
    """

    sql_delete_stored_entry = """
    DELETE FROM 
    stored_entity
    WHERE id IN (SELECT id from se_to_delete)
    """

    sql_delete_tasks = """
    DELETE FROM 
    step_concrete
    WHERE id IN (SELECT id from task_to_delete)
    """

    cursor = taskset._project.persistent_graph.connection.cursor()
    cursor.execute(sql_create_task_tmp)
    cursor.executemany(sql_add_task_id_to_delete, ((task.task_id,) for task in taskset))
    cursor.execute(sql_create_stored_entity_tmp)
    # delete the associations between stored_entities that are the target of tasks in the taskset and tasks
    cursor.execute(sql_delete_target_association)
    # same for sources
    cursor.execute(sql_delete_src_association)
    # delete the stored_entries 
    cursor.execute(sql_delete_stored_entry)
    # delete the tasks in the taskset
    cursor.execute(sql_delete_tasks)
    # delete the directories
    for task in taskset:
        shutil.rmtree(task.dirname)
    cursor.execute("DROP TABLE task_to_delete")
    cursor.execute("DROP TABLE se_to_delete")
    taskset._project.persistent_graph.connection.commit()


class FrozenNamespace(object):
    # Implementation note: a :class:`namedtuple` was not used, because we wanted
    # all class attributes and methods to start with '_', in order to let autocompletion
    # only show instance-defined attributes by default
    def __init__(self, args):
        """ 
        :param args: iterable of (key,value) pairs, the key being the name in the namespace.
        """
        fields = []
        for key, value in args:
            fields.append(key)
            setattr(self, key, value)
        self.__fields = tuple(fields)
    def __setattr__(self, key, value):
        if hasattr(self, key):
            raise ValueError("The attribute '%s' is already defined." % key)
        self.__dict__[key] = value
    @property
    def _fields(self):
        return self.__fields


class Environment(object):
    """
    Represent the current environment in a (presumably) easy way for writing
    recipes.
    """
    def __init__(self, model):
        """ 
        :param model: Python module following the :mod:`railroadtracks` model. 
        """
        classes, activityenum = core.steplist(model), model.ACTIVITY
        stepclasses = list()
        stepinstances = list()
        knownclassnames = set()
        knowninstancenames = set()
        AvailableStepGroup = namedtuple('AvailableStepGroup', 'default_instances classes')
        for cls in unifex._make_stepdict(model).values():
            if not issubclass(cls, core.StepAbstract):
                raise ValueError("Classes used as steps must inherit from core.StepAbstract.")
            stepname = cls.__name__.lower()
            if stepname is None:
                raise ValueError('The step name for class "%s" is not defined.' % cls)
            elif stepname in knownclassnames:
                # check unique class names
                raise ValueError('The step name "%s" is defined twice.' % stepname)
            elif stepname.lower() in knowninstancenames:
                # check unique lower-case for class names (these will be used for instance names)
                raise ValueError('The step name "%s", when converted to lower case, is defined twice.' % stepname)
            stepclasses.append(cls)
            stepinstances.append(cls(cls._default_execpath))
            knownclassnames.add(stepname)
            knowninstancenames.add(stepname.lower())
        
        stepclasses = tuple(stepclasses)
        stepinstances = tuple(stepinstances)

        od = OrderedDict((x, []) for x in activityenum)
        
        for s,si in zip(stepclasses, stepinstances):
            for a in s.activities:
                od[a].append((s, si))

        activitylist = list()
        for activity, classinstancepairs in od.items():
            # iterate over the keys (each being an activity)
            clsnames = list()
            instancenames = list()
            for stepcls, stepinstance in classinstancepairs:
                clsname = stepcls.__name__
                instancename = clsname.lower()
                clsnames.append(clsname)
                instancenames.append(instancename)
            #NT_cls = frozencollection(activity.name+'_cls', instancenames)
            activitylist.append(FrozenNamespace((key, element[1]) for key, element in zip(instancenames, classinstancepairs)))
            #activitylist.append(NT_cls(*(element for element in clsnames)))

        self.__activities = FrozenNamespace((key.name, value) for key, value in zip(od.keys(), activitylist))
        self._stepclasses = stepclasses
        self._stepinstances = stepinstances

    @property
    def activities(self):
        """ Access the activities declared by the model as attributes. """
        return self.__activities

    @property
    def stepclasses(self):
        """ Steps. """
        return self._stepclasses

    @property
    def stepinstances(self):
        """ Default instance for the steps (created from the default executables in the PATH). """
        return self._stepinstances

if __name__ == '__main__':

    def _exec_info(args):
        model = importlib.import_module(args.model)
        project = Project(model, wd=args.wd, db_fn=args.db_fn)
        dbid = DbID(args.task_id)
        res = project.persistent_graph.step_concrete_info(dbid)
        print(res)

    def _exec_cmd(args):
        project = Project(wd=args.wd, db_dn=args.db_fn)
        raise NotImplementedError()

    def _exec_run(args):
        project = Project(wd=args.wd, db_dn=args.db_fn)
        task = project.get_task(args.task_id)
        if task.status == hortator._TASK_DONE:
            print('The task %i is already done' % task.id)
            return
        else:
            try:
                res = task.execute()
                task.status = hortator._TASK_DONE
            except:
                task.status = hortator._TASK_FAILED

    import argparse, importlib
    parser = argparse.ArgumentParser()
    subparsers = parser.add_subparsers(help='Action to perform.')

    parser_info = subparsers.add_parser('status', help="Retrieve information about a step.")
    parser_cmd = subparsers.add_parser('cmd', help="Return the unified execution command line.")
    parser_run = subparsers.add_parser('run', help="Run a step.")

    parser.add_argument('-d', '--working-directory',
                        dest = 'wd',
                        help = 'Working directory for the project.')
    parser.add_argument('-b', '--db-file',
                        dest = 'db_fn',
                        help = 'Database file')
    parser.add_argument('-m', '--model',
                        dest = 'model',
                        help = 'Model (as a Python module)')    
    parser.add_argument(dest = 'task_id',
                        type = int,
                        help = 'Task ID')
    parser_info.set_defaults(func = _exec_info)
    parser_cmd.set_defaults(func = _exec_cmd)
    parser_run.set_defaults(func = _exec_run)


    args = parser.parse_args()
    args.func(args)




# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Differential expression
"""

import os
import argparse
from railroadtracks import core
from railroadtracks import environment
from railroadtracks.model.files import CSVFile, File
from railroadtracks.unifex import _cmdfromuei

RSOURCES_DIR = os.path.dirname(__file__)

class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'DIFFEXP'
    DIFFEXP = 'Differential Expression'

class AssetsDifferentialExpression(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('counttable_fn', CSVFile, ''),
                                          core.AssetAttr('sampleinfo_fn', File, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('diffexp_fn', File, '')])

class DifferentialExpressionMeasurer(core.StepAbstract):
    Assets = AssetsDifferentialExpression
    pass

class RDifferentialExpressionMeasurer(DifferentialExpressionMeasurer):

    activities = (ACTIVITY.DIFFEXP, )
    _default_execpath = 'R'
    parser = None

    def __init__(self, executable = None,
                 rsource_template = None):
        if executable is None:
            executable = type(self)._default_execpath
        if not isinstance(executable, environment.R):
            executable = environment.R(executable)
        self.r = executable
        if rsource_template is None:
            rsource_template = os.path.join(RSOURCES_DIR, 
                                            self._rscript_name)
        assert os.path.isfile(rsource_template), \
            'The needed R script "%s" is not a file' % rsource_template
        self.rsource_template = rsource_template
        self._run_cmd = None
        self._version = None

    _execpath = property(lambda x: x.r.path, None, None)

    @property
    def version(self):
        if self._version is None:
            self._version = self.r.packageversion(self._rpackagename)
        return self._version

    def run(self, assets, parameters=(), magicvariable = 'railroadtracks_import'):
        """ 
        :param sources: named tuple 
        :param targets: name tuple
        """
        assert isinstance(assets, AssetsDifferentialExpression), "The parameter 'assets' must be an %s" % AssetsDifferentialExpression.__name__
            
        var_in = {'counttable_fn': assets.source.counttable_fn.name,
                  'sampleinfo_fn': assets.source.sampleinfo_fn.name,
                  'diffexp_fn': assets.target.diffexp_fn.name}

        if self.parser is not None:
            options, unknown = self.parser.parse_known_args(parameters)
            #FIXME: just ignore unknown arguments ?
            for k,v in options.__dict__.items():
                if k in var_in:
                    raise ValueError("The parameter '%s' is already used internally.")
                var_in[k] = v

        with open(self.rsource_template) as template:
            code = template.read()

        code = os.linesep.join([code, 'run(%s)' % magicvariable])
        returncode = self.r.run_snippet(code, var_in=var_in)
        # FIXME: should return the unified execution command line
        uei = core.UnifiedExecInfo(self.r.path, self._name, assets.source, assets.target, parameters, 
                                   None, None # logging_file and logging_level
        )
        cmd = _cmdfromuei(uei)
        return (cmd, returncode)

# -- note-R-differential-expression-begin

class EdgeR(RDifferentialExpressionMeasurer):
    _rscript_name = 'edger.R'
    _rpackagename = 'edgeR'
    _name = 'edger'

class DESeq(RDifferentialExpressionMeasurer):
    _rscript_name = 'deseq.R'
    _rpackagename = 'DESeq'
    _name = 'deseq'
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--dispersion-fittype',
                        dest = 'dispersion_fittype',
                        choices = ("parametric", "local"),
                        default = "parametric")

class DESeq2(DESeq):
    _rscript_name = 'deseq2.R'
    _rpackagename = 'DESeq2'
    _name = 'deseq2'

class LimmaVoom(RDifferentialExpressionMeasurer):
    _rscript_name = 'limmavoom.R'
    _rpackagename = 'limma'
    _name = 'limma-voom'

# -- note-R-differential-expression-end

class EBSeq(RDifferentialExpressionMeasurer):
    _rscript_name = 'ebseq.R'
    _rpackagename = 'EBSeq'
    _name = 'ebseq'
    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--maxround',
                        dest = 'maxround',
                        help = 'Parameter maxround in EBSeq::EBTest(). Default: %(default)s',
                        type = float,
                        default = 5)
    parser.add_argument('--qtrm',
                        dest = 'qtrm',
                        help = 'Parameter Qtrm in EBSeq::EBTest(). Default: %(default)s',
                        type = float,
                        default = 0.75)
    parser.add_argument('--qtrmcut',
                        dest = 'qtrmcut',
                        help = 'Parameter QtrmCut in EBSeq::EBTest(). Default: %(default)s',
                        type = int,
                        default = 10)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys, os, tempfile, subprocess
import unittest

from railroadtracks import core, rnaseq, hortator
import railroadtracks.model.aligners

if sys.version_info[0] < 3:
    linesep = os.linesep
else:
    linesep = bytes(os.linesep, encoding='ascii')

class ActivitiesSplit(core.Enum):
    HEADTAIL = 'Split in 2'
    MERGE = 'Merge'
    FOO = 'bar'
# Asset for a step that splits a file into head and tail
class AssetsSplit(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('file', core.File, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('head', core.File, ''),
                                          core.AssetAttr('tail', core.File, '')])
# Step that splits a file into head and tail
class Split(core.StepAbstract):
    _name = 'split'
    _default_execpath = None
    Assets=AssetsSplit
    activities = (ActivitiesSplit.HEADTAIL, )
    def __init__(self, executable):
        self._execpath = executable
    @property
    def version(self):
        return '0.1'
    def run(self, assets, parameters=()):
        with open(assets.source.file, 'rb') as fh_in:
            with open(assets.target.head.name, 'wb') as fh_out:
                head = fh_in.read(1)
                fh_out.write(head)
            with open(assets.target.tail.name, 'wb') as fh_out:
                fh_out.write(fh_in)
        cmd = None
        returncode = 0
        return cmd, returncode

class AssetsCat(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('heads', core.FileSequence, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('result', core.File, '')])

class Cat(core.StepAbstract):
    _name = 'cat'
    Assets=AssetsCat
    activities = (ActivitiesSplit.MERGE, )
    _default_execpath = None
    def __init__(self, executable):
        self._execpath = executable
    @property
    def version(self):
        return '0.1'
    def run(self, assets, parameters=()):
        with open(assets.target.result.name, 'wb') as fh_out:
            for x in assets.source.heads:
                with open(x.name, 'rb') as fh_in:
                    fh_out.write(fh_in)
        cmd = None
        returncode = 0
        return cmd, returncode


class StepTestCase(unittest.TestCase):

    def test_Step(self):
        #'id label classname entityname'
        fh = tempfile.NamedTemporaryFile(mode='w+')
        fh.write('foobarbaz')
        fh.flush()
        out_fh = tempfile.NamedTemporaryFile(suffix='.csv')
        src = (hortator.StoredEntity(1,'file', 'File', fh.name), )
        targets = (hortator.StoredEntity(1,'crc', 'CSVFile', out_fh.name), )
        #id status steptype_id executable clsname version
        sc = hortator.StepConcrete_DbEntry(1,
                                           hortator._TASK_TODO,
                                           1,
                                           1, None,
                                           'CRCHeadTail', None, ())
        parameters = tuple()
        step = hortator.Step(sc, src, targets, parameters,
                             rnaseq)
        self.assertFalse(step.iscomplete())
        returncode = step.run()
        self.assertEqual(0, returncode)
        sc = hortator.StepConcrete_DbEntry(1,
                                           hortator._TASK_DONE,
                                           1,
                                           1, None,
                                           'CRCHeadTail', None, ())
        model = core.Model(tuple()) # getting away with an empty model for this step
        step = hortator.Step(sc, src, targets, parameters, model)
        self.assertTrue(step.iscomplete())

class StepGraphTestCase(unittest.TestCase):
    
    def setUp(self):
        self.cache_file = tempfile.NamedTemporaryFile()
        model = core.Model(tuple()) # getting away with an empty model for this step
        self.cache = hortator.PersistentTaskGraph(self.cache_file.name, model, 
                                                 force_create=True)

    @unittest.skip('test not implemented')
    def test_StepGraph(self):
        raise NotImplementedError()
        #FIXME: where are the tests ?

    @unittest.skip('test not implemented')
    def test_addstep(self):
        raise NotImplementedError()
        #FIXME: where are the tests ?

class PersistanceTestCase(unittest.TestCase):

    cls_to_test = hortator.PersistentTaskGraph
    
    def setUp(self):
        self.cache_file = tempfile.NamedTemporaryFile(mode='w+')
        class PythonTime(core.StepAbstract):
            """ Dummy step - print the local date/time using Python. """
            _name = 'time'
            _default_execpath = 'time'
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            def __init__(self, executable=None):
                if executable is None:
                    self._execpath = self._default_execpath
                else:
                    self._execpath = executable
            @property
            def version(self):
                res = subprocess.check_output([self._execpath, '--version'],
                                              stderr=subprocess.STDOUT)
                version = res.split(linesep)[0]
                return version
            def run(self, assets, parameters=('%D', )):
                # assets not used
                res = subprocess.check_output([self._execpath, '-c',
                                               "import %s; print(time.strftime('%D', time.localtime()))" % self._default_execpath])
                print(res)
        self.PythonTime = PythonTime

    def tearDown(self):
        self.cache_file.close()

    def test_PersistentTaskGraph(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        # test the initialization
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        # check that it was created
        self.assertTrue(cache.created)
        # empty set of steps
        s = tuple(cache.iter_steps())
        self.assertEqual(0, len(s))

    def test_statuslist(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        statuslist = cache.statuslist
        self.assertEqual(set((y,x) for x,y in hortator._TASK_STATUS_LIST.items()), set(statuslist))

    def test_reopen(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        #create
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        #reopen
        cache_2 = self.cls_to_test(self.cache_file.name, model)
        # check that it was not created
        self.assertFalse(cache_2.created)
        # check that the statuslist is matching (as it also means that the inserts
        # to set up the DB were committed.
        self.assertEqual(cache.statuslist, cache_2.statuslist)

    def test_id_stored_entity(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        db_id = cache.id_stored_entity(rnaseq.FASTQPossiblyGzipCompressed, 'hohoho.fastq')
        self.assertTrue(db_id.new)
        self.assertEqual(1, db_id.id)
        # same will give the same id.
        db_id_same = cache.id_stored_entity(rnaseq.FASTQPossiblyGzipCompressed, 'hohoho.fastq')
        self.assertFalse(db_id_same.new)
        self.assertEqual(db_id.id, db_id_same.id)
        # different will give a different id.
        db_id_differ = cache.id_stored_entity(rnaseq.FASTQPossiblyGzipCompressed, 'hahaha.fastq')
        self.assertTrue(db_id_differ.new)
        self.assertNotEqual(db_id.id, db_id_differ.id)

    def test_id_stored_sequence_len1(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        db_id = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hohoho.fastq'),))
        self.assertTrue(db_id.new)
        self.assertEqual(1, db_id.id)
        # same will give the same id.
        db_id_same = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hohoho.fastq'),))
        self.assertFalse(db_id_same.new)
        self.assertEqual(db_id.id, db_id_same.id)
        # different will give a different id.
        db_id_differ = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hahaha.fastq'),))
        self.assertTrue(db_id_differ.new)
        self.assertNotEqual(db_id.id, db_id_differ.id)

    def test_id_stored_sequence_len2(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        db_id = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hohoho.fastq'),
                                                             (core.File, 'hahaha.fastq')
                                                         ))
        self.assertTrue(db_id.new)
        self.assertEqual(1, db_id.id)
        # same will give the same id.
        db_id_same = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hohoho.fastq'),
                                                                  (core.File, 'hahaha.fastq')
                                                            ))
        self.assertFalse(db_id_same.new)
        self.assertEqual(db_id.id, db_id_same.id)
        # different will give a different id.
        db_id_differ = cache.id_stored_sequence(core.FileSequence, ((core.File, 'hahaha.fastq'),
                                                                    (core.File, 'hohoho.fastq')
                                                                ))
        self.assertTrue(db_id_differ.new)
        self.assertNotEqual(db_id.id, db_id_differ.id)


    def test_id_step_activity(self):
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        # new activity
        db_id = cache.id_step_activity(rnaseq.ACTIVITY.ALIGN)
        self.assertTrue(db_id.new)
        self.assertEqual(1, db_id.id)
        # other activity - different ID
        db_id_other = cache.id_step_activity(rnaseq.ACTIVITY.QUANTIFY)
        self.assertTrue(db_id_other.new)
        self.assertNotEqual(db_id.id, db_id_other.id)
        db_id_another = cache.id_step_activity(rnaseq.ACTIVITY.DIFFEXP)
        self.assertTrue(db_id_another.new)
        self.assertNotEqual(db_id.id, db_id_another.id)
        # already know activity - do not create a new one
        db_id_same = cache.id_step_activity(rnaseq.ACTIVITY.ALIGN)
        self.assertFalse(db_id_same.new)
        self.assertEqual(db_id.id, db_id_same.id)

    def test_id_step_type_oneactivity(self):
        # step_type qualifies the activites carried out by a step
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        # new step_type that does only alignments
        db_id = cache.id_step_type((rnaseq.ACTIVITY.ALIGN, ))
        self.assertTrue(db_id.new)
        self.assertEqual(1, db_id.id)
        # step_type that does only alignments
        db_id = cache.id_step_type((rnaseq.ACTIVITY.ALIGN, ))
        self.assertFalse(db_id.new)
        self.assertEqual(1, db_id.id)

    def test_id_step_type_severalactivities(self):
        # step_type qualifies the activites carried out by a step
        cache = self.cls_to_test(self.cache_file.name, rnaseq, force_create=True)
        # new step_type that does pretty much anything
        db_id_all = cache.id_step_type((rnaseq.ACTIVITY.ALIGN,
                                        rnaseq.ACTIVITY.QUANTIFY,
                                        rnaseq.ACTIVITY.DIFFEXP))
        self.assertTrue(db_id_all.new)
        self.assertEqual(1, db_id_all.id)
        # same step_type - indedependent from the order in which activities are specified
        # (this is by design)
        db_id_all_same = cache.id_step_type((rnaseq.ACTIVITY.ALIGN,
                                             rnaseq.ACTIVITY.DIFFEXP,
                                             rnaseq.ACTIVITY.QUANTIFY))
        self.assertFalse(db_id_all_same.new)
        self.assertEqual(db_id_all.id, db_id_all_same.id)

    def test_id_step_variant(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class Activities(core.Enum):
            COMPRESS = 'Compress'
        class GZip(core.StepAbstract):
            _name = 'gzip'
            _default_execpath = 'gzip'
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            def __init__(self, executable=None):
                if executable is None:
                    self._execpath = _default_execpath
                else:
                    self._execpath = executable
            @property
            def version(self):
                res = subprocess.check_output([self._execpath, '--version'])
                version = res.split(linesep)[0]
                return version
            def run(self):
                pass
        gzip = GZip('gzip')
        db_id = cache.id_step_variant(gzip,
                                      (Activities.COMPRESS,))
        self.assertTrue(isinstance(db_id.id, int))
        #FIXME: dummy executable (no sources or targets)

    def test_id_step_variant_noexecutable(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class Activities(core.Enum):
            NOODLE = 'Noodle'
        class Foo(core.StepAbstract):
            _name = 'foo'
            _default_execpath = None
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            Assets = None
            def __init__(self, executable):
                if executable is not None:
                    raise ValueError('No executable needed. Should be None.')
                self._execpath = None
            @property
            def version(self):
                return '0.1.0'
            def run(self):
                pass
        foo = Foo(None)
        db_id = cache.id_step_variant(foo,
                                      (Activities.NOODLE,))
        self.assertTrue(db_id.new)
        self.assertTrue(isinstance(db_id.id, int))
        db_id_same = cache.id_step_variant(foo,
                                           (Activities.NOODLE,))
        self.assertFalse(db_id_same.new)
        self.assertEqual(db_id.id, db_id_same.id)
        
    def test_id_stepconcrete(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class Activities(core.Enum):
            DATETIME = 'Give date/time'
        PythonTime = self.PythonTime

        python = PythonTime('python')
        stepvariant_db_id = cache.id_step_variant(python,
                                                  (Activities.DATETIME,))
        # empty sources is a special case
        sources = core.AssetSet() # source
        targets = core.AssetSet() # targets
        parameters = tuple()
        db_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                      sources, targets, parameters)
        db_id_same = cache.id_stepconcrete(stepvariant_db_id.id,
                                           sources, targets, parameters)
        self.assertEqual(db_id.id, db_id_same.id)
        db_id_notthesame = cache.id_stepconcrete(stepvariant_db_id.id,
                                                 sources, targets, parameters, tag = 2)
        self.assertNotEqual(db_id.id, db_id_notthesame.id)

        # 1-element sources
        sources = railroadtracks.model.aligners.AssetsIndexer.Source(rnaseq.FASTAFile('foo.fasta'))

        db_id_nothesame = cache.id_stepconcrete(stepvariant_db_id.id,
                                                sources, targets, parameters)
        self.assertNotEqual(db_id.id, db_id_nothesame.id)
        db_id_sameagain = cache.id_stepconcrete(stepvariant_db_id.id,
                                                sources, targets, parameters)
        
        self.assertEqual(db_id_sameagain.id, db_id_nothesame.id)
        db_id_nothesameagain = cache.id_stepconcrete(stepvariant_db_id.id,
                                                     sources, targets, ("%Y",))
        self.assertNotEqual(db_id.id, db_id_nothesameagain.id)
        self.assertNotEqual(db_id_sameagain.id, db_id_nothesameagain.id)

        # 1-element sources, several parameters
        db_id_2params = cache.id_stepconcrete(stepvariant_db_id.id,
                                              sources, targets, ("%Y", "Z"))
        db_id_same2params = cache.id_stepconcrete(stepvariant_db_id.id,
                                                  sources, targets, ("%Y", "Z"))
        self.assertEqual(db_id_2params.id, db_id_same2params.id)

        db_id_2otherparams = cache.id_stepconcrete(stepvariant_db_id.id,
                                                   sources, targets, ("%Y", "W"))
        self.assertNotEqual(db_id_2params.id, db_id_2otherparams.id)

        # 2-elements sources
        SrcCls = core.assetfactory('Source', 
                                   [core.AssetAttr('reference', rnaseq.FASTAFile, ''),
                                    core.AssetAttr('otherreference', rnaseq.FASTAFile, '')])
        sources = SrcCls(rnaseq.FASTAFile('foo.fasta'),
                         rnaseq.FASTAFile('bar.fasta'))
        db_id_notthesame = cache.id_stepconcrete(stepvariant_db_id.id,
                                                 sources, targets, parameters)
        self.assertNotEqual(db_id.id, db_id_notthesame.id)
        db_id_sameagain = cache.id_stepconcrete(stepvariant_db_id.id,
                                                sources, targets, parameters)
        self.assertEqual(db_id_sameagain.id, db_id_notthesame.id)


        # 1-element source / 1-element target
        sources = railroadtracks.model.aligners.AssetsIndexer.Source(rnaseq.FASTAFile('foo.fasta'))
        targets = railroadtracks.model.aligners.AssetsIndexer.Target(rnaseq.FilePattern('foo_idx'))

        foo_sh = rnaseq.Anyscript()
        stepvariant_db_id = cache.id_step_variant(foo_sh,
                                                  (Activities.DATETIME,))

        db_id_nothesame = cache.id_stepconcrete(stepvariant_db_id.id,
                                                sources, targets, parameters)
        self.assertNotEqual(db_id.id, db_id_nothesame.id)
        db_id_sameagain = cache.id_stepconcrete(stepvariant_db_id.id,
                                                sources, targets, parameters)
        self.assertEqual(db_id_sameagain.id, db_id_nothesame.id)

        # fail if target assets are suddenly different
        targets_bar = railroadtracks.model.aligners.AssetsIndexer.Target(rnaseq.FilePattern('bar_idx'))
        self.assertRaises(ValueError, cache.id_stepconcrete, stepvariant_db_id.id,
                          sources, targets_bar, parameters)


    def test_get_assets(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class Activities(core.Enum):
            DATETIME = 'Give date/time'
        PythonTime = self.PythonTime
        python = PythonTime('python')
        stepvariant_db_id = cache.id_step_variant(python,
                                                  (Activities.DATETIME,))
        # 2-elements sources
        SrcCls = core.assetfactory('Source', 
                                   [core.AssetAttr('reference', rnaseq.FASTAFile, ''),
                                    core.AssetAttr('otherreference', rnaseq.FASTAFile, ''),
                                    core.AssetAttr('listoffiles', rnaseq.CSVFileSequence, '')])
        sources = SrcCls(rnaseq.FASTAFile('foo.fasta'),
                         rnaseq.FASTAFile('bar.fasta'),
                         rnaseq.CSVFileSequence((rnaseq.CSVFile('baz.csv'),rnaseq.CSVFile('baz2.csv'))))
        targets = core.AssetSet() # targets
        parameters = tuple()
        db_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                      sources, targets, parameters)
        for storedthing in cache.get_srcassets(db_id.id):
            thing = storedthing.resurrect(rnaseq)

    def test_nconcrete_steps(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class Activities(core.Enum):
            DATETIME = 'Give date/time'
        PythonTime = self.PythonTime
        python = PythonTime('python')
        stepvariant_db_id = cache.id_step_variant(python,
                                                  (Activities.DATETIME,))
        sources = core.AssetSet() # source
        targets = core.AssetSet() # targets
        parameters = tuple()
        self.assertEqual(0, cache.nconcrete_steps)
        taskstatuscount = cache.nconcrete_steps_status
        self.assertEqual(0, len(taskstatuscount))
        db_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                      sources, targets, parameters)
        self.assertEqual(1, cache.nconcrete_steps)
        taskstatuscount = cache.nconcrete_steps_status
        self.assertEqual(taskstatuscount[0].label, hortator._TASK_TODO)
        self.assertEqual(taskstatuscount[0].count, 1)


    def test_gettargetsofactivity(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        split = Split(None)
        parameters = tuple()
        input_file = tempfile.NamedTemporaryFile(mode='w+')
        input_file.write('123')
        input_file.flush()
        head_file = tempfile.NamedTemporaryFile(mode='w+')
        tail_file = tempfile.NamedTemporaryFile(mode='w+')
        sources = split.Assets.Source(core.File(input_file.name))
        targets = split.Assets.Target(core.File(head_file.name),
                                      core.File(tail_file.name))
        self.assertRaises(ValueError, cache.get_targetsofactivity, ActivitiesSplit.HEADTAIL)
        # ensure that step variant is tracked
        stepvariant_db_id = cache.id_step_variant(split,
                                                  split.activities)
        self.assertRaises(ValueError, cache.get_targetsofactivity, ActivitiesSplit.FOO)
        # get the targets of the activity (obviously there are not any yet)
        res = cache.get_targetsofactivity(ActivitiesSplit.HEADTAIL)
        self.assertEqual(0, len(res))
        # create a new task
        task_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                        sources, targets, parameters)
        # there is only one such 
        res = cache.get_targetsofactivity(ActivitiesSplit.HEADTAIL)
        self.assertEqual(2, len(res))
        #
        head_file = tempfile.NamedTemporaryFile(mode='w+')
        tail_file = tempfile.NamedTemporaryFile(mode='w+')
        targets_other = split.Assets.Target(core.File(head_file.name),
                                      core.File(tail_file.name))
        task_id_other = cache.id_stepconcrete(stepvariant_db_id.id,
                                              sources, targets_other, parameters=(1,))
        res = cache.get_targetsofactivity(ActivitiesSplit.HEADTAIL)
        self.assertEqual(4, len(res))
        # --
        cat = Cat(None)
        concat = tempfile.NamedTemporaryFile(mode='w+')
        sources = cat.Assets.Source(core.FileSequence((targets.head, targets_other.head)))
        targets = cat.Assets.Target(core.File(concat.name))
        # ensure that step variant is tracked
        stepvariant_db_id = cache.id_step_variant(cat,
                                                  cat.activities)
        task_id_cat = cache.id_stepconcrete(stepvariant_db_id.id,
                                            sources, targets, parameters)
        res = cache.get_targetsofactivity(ActivitiesSplit.MERGE)
        self.assertEqual(1, len(res))
                                        
    def test_get_parenttask_of_storedentity(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        split = Split(None)
        parameters = tuple()
        input_file = tempfile.NamedTemporaryFile(mode='w+')
        input_file.write('123')
        input_file.flush()
        head_file = tempfile.NamedTemporaryFile(mode='w+')
        tail_file = tempfile.NamedTemporaryFile(mode='w+')
        sources = split.Assets.Source(core.File(input_file.name))
        targets = split.Assets.Target(core.File(head_file.name),
                                      core.File(tail_file.name))
        # ensure that step variant is tracked
        stepvariant_db_id = cache.id_step_variant(split,
                                                  split.activities)
        # create a new task
        task_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                        sources, targets, parameters)
        for sa in cache.get_srcassets(task_id):
            # all are root nodes
            res = cache.get_parenttask_of_storedentity(sa)
            self.assertTrue(res is None)
        for sa in cache.get_targetassets(task_id):
            res = cache.get_parenttask_of_storedentity(sa)
            self.assertEqual(task_id.id, res.id)
        #
        head_file_other = tempfile.NamedTemporaryFile(mode='w+')
        tail_file_other = tempfile.NamedTemporaryFile(mode='w+')
        targets_other = split.Assets.Target(core.File(head_file_other.name),
                                            core.File(tail_file_other.name))
        task_id_other = cache.id_stepconcrete(stepvariant_db_id.id,
                                              split.Assets.Source(targets.tail), targets_other, parameters=(1,))
        for sa in cache.get_srcassets(task_id_other):
            res = cache.get_parenttask_of_storedentity(sa)
            self.assertEqual(task_id.id, res.id)
        for sa in cache.get_targetassets(task_id_other):
            res = cache.get_parenttask_of_storedentity(sa)
            self.assertEqual(task_id_other.id, res.id)

        # --
        cat = Cat(None)
        sources_cat = cat.Assets.Source(core.FileSequence((targets.head, targets_other.head)))
        concat = tempfile.NamedTemporaryFile(mode='w+')
        targets_cat = cat.Assets.Target(core.File(concat.name))
        # ensure that step variant is tracked
        stepvariant_db_id_cat = cache.id_step_variant(cat,
                                                      cat.activities)
        task_id_cat = cache.id_stepconcrete(stepvariant_db_id_cat.id,
                                            sources_cat, targets_cat, parameters)
        for sa in cache.get_srcassets(task_id_cat):
            if hasattr(sa, 'iter_storedentities'):
                for sa_sub, t in zip(sa.iter_storedentities(), (task_id, task_id_other)):
                    res = cache.get_parenttask_of_storedentity(sa_sub)
                    self.assertEqual(t.id, res.id)                    
            else:
                res = cache.get_parenttask_of_storedentity(sa)
                self.assertEqual(task_id_cat.id, res.id)

        head_file_other = tempfile.NamedTemporaryFile(mode='w+')
        tail_file_other = tempfile.NamedTemporaryFile(mode='w+')
        targets_other = split.Assets.Target(core.File(head_file_other.name),
                                            core.File(tail_file_other.name))
        task_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                        split.Assets.Source(targets_cat.result), targets_other, tuple())
        for sa in cache.get_srcassets(task_id):
            res = cache.get_parenttask_of_storedentity(sa)
            self.assertEqual(task_id_cat.id, res.id)
        


    def test_gettargetsoftype(self):
        model = core.Model(tuple()) # getting away with an empty model for this step
        cache = self.cls_to_test(self.cache_file.name, model, force_create=True)
        class SplitCSV(Split):
            _name = 'split'
            activities = (core.DEFAULT_ACTIVITY.MISC, )
            version = '0.1'
            _default_execpath = 'None'
            class Assets(core.AssetsStep):
                Source = core.assetfactory('Source', [core.AssetAttr('file', railroadtracks.model.files.CSVFile, '')])
                Target = core.assetfactory('Target', [core.AssetAttr('head', railroadtracks.model.files.CSVFile, ''),
                                                      core.AssetAttr('tail', railroadtracks.model.files.CSVFile, '')])
            def run(self, assets, parameters=()):
                with open(assets.source.file, 'rb') as fh_in:
                    csv_r = csv.reader(fh_in)
                    with open(assets.target.head, 'wb') as fh_out:
                        csv_w = csv.writer(fh_out)
                        head = next(csv_r)
                        csv_w.writerow(head)
                    with open(assets.target.tail, 'wb') as fh_out:
                        csv_w = csv.writer(fh_out)
                        for row in csv_r:
                            csv_w.writerow(row)
                cmd = None
                returncode = 1
                return cmd, returncode

        split = Split(None)
        parameters = tuple()
        input_file = tempfile.NamedTemporaryFile(mode='w+')
        input_file.write('123')
        input_file.flush()
        head_file = tempfile.NamedTemporaryFile(mode='w+')
        tail_file = tempfile.NamedTemporaryFile(mode='w+')
        sources = split.Assets.Source(core.File(input_file.name))
        targets = split.Assets.Target(core.File(head_file.name),
                                      core.File(tail_file.name))
        stepvariant_db_id = cache.id_step_variant(split,
                                                  split.activities)
        res = cache.get_targetsoftype(core.File.__name__)
        self.assertEqual(0, len(res))
        task_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                        sources, targets, parameters)
        res = cache.get_targetsoftype(core.File.__name__)
        self.assertEqual(2, len(res))

        splitcsv = SplitCSV(None)
        parameters = tuple()
        input_file = tempfile.NamedTemporaryFile(suffix=".csv", mode='w+')
        input_file.write('123')
        input_file.flush()
        head_file = tempfile.NamedTemporaryFile(suffix=".csv", mode='w+')
        tail_file = tempfile.NamedTemporaryFile(suffix=".csv", mode='w+')
        sources = splitcsv.Assets.Source(railroadtracks.model.files.CSVFile(input_file.name))
        targets = splitcsv.Assets.Target(railroadtracks.model.files.CSVFile(head_file.name),
                                         railroadtracks.model.files.CSVFile(tail_file.name))
        stepvariant_db_id = cache.id_step_variant(splitcsv,
                                                  splitcsv.activities)
        res = cache.get_targetsoftype(railroadtracks.model.files.CSVFile.__name__)
        self.assertEqual(0, len(res))
        task_id = cache.id_stepconcrete(stepvariant_db_id.id,
                                        sources, targets, parameters)
        res = cache.get_targetsoftype(railroadtracks.model.files.CSVFile.__name__)
        self.assertEqual(2, len(res))

        head_file = tempfile.NamedTemporaryFile(mode='w+')
        tail_file = tempfile.NamedTemporaryFile(mode='w+')
        targets = split.Assets.Target(core.File(head_file.name),
                                      core.File(tail_file.name))
        task_id_other = cache.id_stepconcrete(stepvariant_db_id.id,
                                              sources, targets, parameters=(1,))
        res = cache.get_targetsoftype(core.File.__name__)
        self.assertEqual(4, len(res))


class CachedPersistanceTestCase(PersistanceTestCase):

    cls_to_test = hortator.CachedPersistentTaskGraph



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import sys
from abc import ABCMeta, abstractproperty
import os
import logging
logger = logging.getLogger(__name__)
import subprocess
import re
import tempfile
import shutil
import argparse
import warnings
from railroadtracks.unifex import UnifexError
from railroadtracks import (core,
                            environment)
from railroadtracks.model.files import (FilePattern,
                                        FASTQPossiblyGzipCompressed,
                                        FASTAFile,
                                        SAMFile,
                                        BAMFile,
                                        samtools_getversion,
                                        SamtoolsSamToBam)


class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'INDEX ALIGN UTILITY'
    INDEX = 'Index'
    ALIGN = 'Align'
    UTILITY = 'Utility'


class AssetsIndexer(core.AssetsStep):
    """
    Assets for a general indexing step.
    """
    Source = core.assetfactory('Source', 
                               [core.AssetAttr('reference',
                                               FASTAFile, '')])
    Target = core.assetfactory('Target', 
                               [core.AssetAttr('indexfilepattern',
                                               FilePattern, '')])

class IndexerAbstract(core.StepAbstract):
    """
    Parent class for an indexing step
    """
    __metaclass__ = ABCMeta
    activities = (ACTIVITY.INDEX, )
    Assets = AssetsIndexer


class BWAIndexFiles(FilePattern):
    """ Set of BWA index files. """
    _extension = ('.fai', '.rpac', '.amb', '.ann', '.pac', '.bwt', '.rbwt', '.rsa', '.sa')

class SavedBowtieIndex(FilePattern):
    """ Set of Bowtie index files. """
    _extension = ('.ebwt', )


class BowtieBuild(IndexerAbstract):
    """
    Bowtie as an indexer.
    
    Should the way it is run on the command line change, a child class
    will be implemented (and the relevant methods be overriden).
    """
    #FIXME: Bowtie2Build inherit from BowtieBuild ?
    _name = 'bowtie-build'
    _default_execpath = 'bowtie-build'
    class Assets(AssetsIndexer):
        Target = core.assetfactory('Target',
                                   [core.AssetAttr('indexfilepattern', 
                                                   SavedBowtieIndex, '')])

    def __init__(self, executable=None):
        """        
        :param executable: path to the executable
        :type param: :class:`str`
        """
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                tmp = subprocess.check_output(cmd)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))
            m = re.match(b'^.+? version ([^ \n]+).*', tmp)
            self._version = m.groups()[0]
        return str(self._version)

    def run(self, assets, parameters=tuple()):
        """ The step on the assets given as parameters. 
        :param assets: the assets to run the step with.
        :param parameters: optional parameters as a sequence at the exception of the assets
        (e.g. ('-k 24', '--super-fast') ).
        :type parameters: a sequence. """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' must be an %s" % core.AssetsStep.__name__
        basename_index = assets.target.indexfilepattern.name
        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        cmd.extend(['-f',
                    assets.source.reference.name, 
                    assets.target.indexfilepattern.name])
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)

        #FIXME: test return code ?
        return (cmd, returncode)


class SailfishIndex(IndexerAbstract):
    _name = 'sailfish-index'
    _default_execpath = 'sailfish'

    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--kmerSize',
                        dest = 'kmersize',
                        required = True,
                        type = int)
    parser.add_argument('--threads',
                        type = int,
                        default = 2)
    #FIXME: duplication with parameter declaration above. should be a way to refactor this.
    PARAMETERS_DEFAULT = ('--kmerSize', '16', '--threads', '2')

    def __init__(self, executable=None):
        """
        :param executable: the executable is R. If None, the class-level
        attribute :attr:`_default_execpath` will be used.
        :type executable: a :class:`str` or a :class:`environment.R`
        """
        if type(self) == SailfishIndex:
            warnings.warn('Model SailfishIndex is deprecated. Please use SalmonIndex.')
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                tmp = subprocess.check_output(cmd, stderr=subprocess.STDOUT)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))
            m = re.match(b'^version : (.+)', tmp)
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters = PARAMETERS_DEFAULT):
        """ 
        :param assets: 
        :type assests: instance of :class:`core.AssetsStep` (or child class)
        :param targets:
        :type parameters: :class:`tuple`
        """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, self.Assets), \
            "The parameter 'assets' must be a '%s', not a '%s'" % (self.Assets.__name__,
                                                                   type(assets).__name__)
        options, unknown = self.parser.parse_known_args(parameters)
        basename_index = assets.target.indexfilepattern.name
        cmd = [self._execpath, 'index']
        cmd.extend(parameters)
        cmd.extend(['--transcripts',
                    assets.source.reference.name, 
                    '--out',
                    assets.target.indexfilepattern.name])
        
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
        return (cmd, returncode)


class SalmonIndex(SailfishIndex):
    _name = 'salmon-index'
    _default_execpath = 'salmon'

    parser = argparse.ArgumentParser(_name)
    parser.add_argument('--threads',
                        type = int,
                        default = 2)

    def run(self, assets, parameters = ()):
        """ 
        :param assets: 
        :type assests: instance of :class:`core.AssetsStep` (or child class)
        :param targets:
        :type parameters: :class:`tuple`
        """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, self.Assets), \
            "The parameter 'assets' must be a '%s', not a '%s'" % (self.Assets.__name__,
                                                                   type(assets).__name__)
        options, unknown = self.parser.parse_known_args(parameters)
        basename_index = assets.target.indexfilepattern.name
        cmd = [self._execpath, 'index']
        cmd.extend(parameters)
        cmd.extend(['--transcripts',
                    assets.source.reference.name, 
                    '-i',
                    assets.target.indexfilepattern.name])
        
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
        return (cmd, returncode)


def _bwa_version(execpath):
    cmd = [execpath, ]
    m = None
    logfile = tempfile.NamedTemporaryFile()
    with open(os.devnull, 'w') as fnull:
        try:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd,
                                         stderr=logfile,
                                         stdout=logfile )
            # BWA is returning 1
            assert returncode==1, "BWA should have been failing with return code 1 (and it did not)"
            # now dig the information out (note: the subprocess has closed the file - open it again)
            with open(logfile.name, mode='rb') as logfile:
                for row in logfile:
                    m = re.match(b'^Version: (.+)$', row)
                    if m is not None:
                        break
        except OSError as ose:
            raise UnifexError("""Command: %s
            %s""" % (' '.join(cmd), ose))
    if m is None:
        raise ValueError('The version number could be determined.')
    return m.groups()[0]


class BWAIndex(IndexerAbstract):
    """
    BWA as an indexer.
    
    """
    _name = 'bwa-index'
    _default_execpath = 'bwa'
    class Assets(AssetsIndexer):
        Target = core.assetfactory('Target',
                                   [core.AssetAttr('indexfilepattern', 
                                                   BWAIndexFiles, '')])

    def __init__(self, executable=None):
        """        
        :param executable: path to the executable
        :type param: :class:`str`
        """
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = _bwa_version(self._execpath)
        return self._version

    def run(self, assets, parameters=tuple()):
        """ The step on the assets given as parameters. 
        :param assets: the assets to run the step with.
        :param parameters: optional parameters as a sequence at the exception of the assets
        :type parameters: a sequence. """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' must be an %s" % core.AssetsStep.__name__
        basename_index = assets.target.indexfilepattern.name
        cmd = ['%s' % self._execpath]
        cmd.append('index')
        cmd.extend(parameters)
        cmd.extend(['-p',
                    assets.target.indexfilepattern.name,
                    assets.source.reference.name])
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)

        #FIXME: test return code ?
        return (cmd, returncode)


class SavedBowtie2Index(FilePattern):
    """ Set of Bowtie2 index files. """
    _extension = ('.bt2', '.bt21')


class Bowtie2Build(IndexerAbstract):
    """
    Bowtie2 as an indexer.
    
    Should the way it is run on the command line change, a child class
    will be implemented (and the relevant methods be overriden).
    """

    _name = 'bowtie2-build'
    _default_execpath = 'bowtie2-build'
    class Assets(AssetsIndexer):
        Target = core.assetfactory('Target',
                                   [core.AssetAttr('indexfilepattern', 
                                                   SavedBowtie2Index, '')])

    def __init__(self, executable=None):
        """        
        :param executable: path to the executable
        :type param: :class:`str`
        """
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                tmp = subprocess.check_output(cmd)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))
            m = re.match(b'^.+? version ([^ \n]+).*', tmp)
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        """ The step on the assets given as parameters. 
        :param assets: the assets to run the step with.
        :param parameters: optional parameters as a sequence at the exception of the assets
        (e.g. ('-k 24', '--super-fast') ).
        :type parameters: a sequence. """
        assert environment.Executable.ispresent(self._execpath), "The executable %s could not be found in the PATH" % self._execpath
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' must be an %s" % core.AssetsStep.__name__
        basename_index = assets.target.indexfilepattern.name
        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        cmd.extend(['-f',
                    assets.source.reference.name, 
                    assets.target.indexfilepattern.name])
        with open(os.devnull, "w") as fnull:
            logging.debug(cmd)
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)

        #FIXME: test return code ?
        return (cmd, returncode)


class AssetsAligner(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', FilePattern, ''),
                                          core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                          core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])
    Target = core.assetfactory('Target', [core.AssetAttr('alignment', BAMFile, '')])


class AlignerAbstract(core.StepAbstract):
    """

    """
    __metaclass__ = ABCMeta
    activities = (ACTIVITY.ALIGN, )
    Assets = AssetsAligner

class Bowtie2(AlignerAbstract):
    """
    Bowtie2 as an aligner.
    
    Should the way it is run on the command line change, a child class should
    be implemented (and the relevant methods be overriden).

    bt = Bowtie2(execpath)
    
    :param execpath: path to the executable

    """

    _name = 'bowtie2'
    _default_execpath = 'bowtie2'
    # vanilla aligner:
    # Source = assetfactory('Source', [core.AssetAttr('indexfilepattern', FilePattern, ''),
    #                                  core.AssetAttr('read1', SavedFASTQ, ''),
    #                                  core.AssetAttr('read2', SavedFASTQ, '', allownone=True)])
    class Assets(AssetsAligner):
        Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', SavedBowtie2Index, ''),
                                              core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                res = subprocess.check_output(cmd)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))

            m = re.match(b'^.+? version ([^ \n]+).*', res)
            if m is None:
                raise RuntimeError('Could not find the version number.')
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        # samtools is a dependency. In the absence of features in RRT that would handle secondary depencies,
        # there is a simple assert here to prevent cryptic failures if the software is not on the system.
        samtools = 'samtools'
        assert environment.Executable.ispresent(self._execpath)
        assert environment.Executable.ispresent(samtools)

        source = assets.source
        # FIXME: these checks should be moved to the Assets class !
        index_pattern = re.compile('^.+\.[1-9][0-9]*\.bt2$')
        indexfiles = [name for cls, name in source.indexfilepattern.iterlistfiles() if index_pattern.match(name)]
        if len(indexfiles) == 0:
            raise ValueError("No bowtie2 index files in %s" % source.indexfilepattern)

        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        if source.read2 is None:
            # single reads
            cmd_sub = ('-U %s' % source.read1.name, )
        else:
            cmd_sub = ('-1', source.read1.name, '-2', source.read2.name)


        unaligned_fn = None
        aligned_fn = None

        # for ext in sorted(assets.target.alignment._extension, key=len, reverse=True):
        #     if assets.target.alignment.name.endswith(ext):
        #         fn = assets.target.alignment.name
        #         fn = fn[:(len(fn)-len(ext))]
        #         unaligned_fn = fn + '_unaligned' + ext
        #         aligned_fn = fn + '_aligned' + ext
        #         break
        # if unaligned_fn is None:
        #     logger.error('Could not build a file name for unaligned reads.')
        #     #FIXME: abort here.

        # if source.read2 is None:
        #     if os.path.exists(unaligned_fn):
        #         logger.warn('File %s already existing.' % unaligned_fn)
        # else:
        #     # "ext" obtained from loop above
        #     for r_i in ('1', '2'):
        #         tmp_fn = re.sub('(.+)(\\%s)$' % ext, '\\1_%s\\2' % r_i, unaligned_fn)
        #         if os.path.exists(tmp_fn):
        #             logger.warn('File %s already existing.' % unaligned_fn)

        # build command line
        # notes:
        #    - q : FASTQ in input
        cmd_align = ['%s' % self._execpath]
        cmd_align.extend(parameters)
        #cmd_align.extend(('--un', unaligned_fn))
        cmd_align.extend(('-x', source.indexfilepattern.name, 
                    '-q'))
        cmd_align.extend(cmd_sub)

        cmd_sam2bam = (samtools, 'view', '-bS', '-')

        aligned_fn = assets.target.alignment.name
        cmd = cmd_align + ['|', ] + list(cmd_sam2bam) + ['>',] + [aligned_fn, ] 
        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull, \
             open(aligned_fn, "w") as fh_out, \
             core.Popen(cmd_align, stdout=subprocess.PIPE, stderr=fnull) as p_align:
            # we are done yet: the output should be BAM, not SAM
            with core.Popen(cmd_sam2bam, 
                            stdin=p_align.stdout, 
                            stdout=fh_out, 
                            stderr=fnull) as p_sam2bam:
                p_sam2bam.communicate()[0]
                returncode = p_sam2bam.returncode

        # now merge aligned and unaligned reads
        #cmd_mergebam = (samtools, 'merge', assets.target.alignment.name, aligned_fn, unaligned_fn)
        #try:
        #    subprocess.check_call(cmd_mergebam)
        #finally:
        #    os.unlink(aligned_fn)
        #    os.unlink(unaligned_fn)
        return (cmd, returncode)

class Bowtie(AlignerAbstract):
    """
    Bowtie as an aligner.
    
    Should the way it is run on the command line change, a child class should
    be implemented (and the relevant methods be overriden).

    bt = Bowtie(execpath)
    
    :param execpath: path to the executable

    """

    _name = 'bowtie'
    _default_execpath = 'bowtie'

    class Assets(AssetsAligner):
        Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', SavedBowtieIndex, ''),
                                              core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                res = subprocess.check_output(cmd)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))

            m = re.match(b'^.+? version ([^ \n]+).*', res)
            if m is None:
                raise RuntimeError('Could not find the version number.')
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        # samtools is a dependency. In the absence of features in RRT that would handle secondary depencies,
        # there is a simple assert here to prevent cryptic failures if the software is not on the system.
        samtools = 'samtools'
        assert environment.Executable.ispresent(self._execpath)
        assert environment.Executable.ispresent(samtools)
        
        source = assets.source
        indexfiles = tuple(name for cls, name in source.indexfilepattern.iterlistfiles())
        if len(indexfiles) == 0:
            raise ValueError("No bowtie index files in %s" % source.indexfilepattern)

        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        if source.read2 is None:
            # single reads
            cmd_sub = (source.read1.name, )
        else:
            cmd_sub = ('-1', source.read1.name, '-2', source.read2.name)

        # Gzip-compression not supported by bowtie1
        if source.read1.iscompressed or (source.read2 is not None and source.read2.iscompressed):
            raise NotImplementedError("Bowtie(1) does not support gzip-compressed FASTQ file. On-the-fly might eventually be implemented in the future.")
        # build command line
        # notes:
        #    - q : FASTQ in input
        cmd_align = ['%s' % self._execpath]
        cmd_align.extend(parameters)
        # unaligned_fn = None
        # aligned_fn = None
        # for ext in sorted(assets.target.alignment._extension, key=len, reverse=True):
        #     if assets.target.alignment.name.endswith(ext):
        #         fn = assets.target.alignment.name
        #         fn = fn[:(len(fn)-len(ext))]
        #         unaligned_fn = fn + '_unaligned' + ext
        #         aligned_fn = fn + '_aligned' + ext
        #         break
        # if unaligned_fn is None:
        #     logger.error('Could not build a file name for unaligned reads.')
        #     #FIXME: abort here.

        # if source.read2 is None:
        #     if os.path.exists(unaligned_fn):
        #         logger.warn('File %s already existing.' % unaligned_fn)
        # else:
        #     # "ext" obtained from loop above
        #     for r_i in ('1', '2'):
        #         tmp_fn = re.sub('(.+)(\\%s)$' % ext, '\\1_%s\\2' % r_i, unaligned_fn)
        #         if os.path.exists(tmp_fn):
        #             logger.warn('File %s already existing.' % unaligned_fn)

        # cmd_align.extend(('--un', unaligned_fn,))
        cmd_align.append('--sam') # otherwise a SAM-like format is produced...
        cmd_align.extend([source.indexfilepattern.name, 
                          '-q'])
        cmd_align.extend(cmd_sub)

        cmd_sam2bam = (samtools, 'view', '-bS', '-')

        aligned_fn = assets.target.alignment.name
        cmd = cmd_align + ['|', ] + list(cmd_sam2bam) + ['>',] + [aligned_fn, ] 
        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull, \
             open(aligned_fn, "w") as fh_out, \
             core.Popen(cmd_align, stdout=subprocess.PIPE, stderr=fnull) as p_align:
            # we are done yet: the output should be BAM, not SAM
            with core.Popen(cmd_sam2bam,
                            stdin=p_align.stdout,
                            stdout=fh_out,
                            stderr=fnull) as p_sam2bam:
                p_align.stdout.close()
                p_sam2bam.communicate()[0]
                returncode = p_sam2bam.returncode

        # # now merge aligned and unaligned reads
        # cmd_mergebam = (samtools, 'merge', assets.target.alignment.name, aligned_fn, unaligned_fn)
        # try:
        #     subprocess.check_call(cmd_mergebam)
        # finally:
        #     os.unlink(aligned_fn)
        #     os.unlink(unaligned_fn)

        return (cmd, returncode)


class BWA(AlignerAbstract):
    """
    BWA as an aligner.
    
    Should the way it is run on the command line change, a child class should
    be implemented (and the relevant methods be overriden).

    bt = Bowtie(execpath)
    
    :param execpath: path to the executable

    """

    _name = 'bwa-align'
    _default_execpath = 'bwa'

    class Assets(AssetsAligner):
        Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', BWAIndexFiles, ''),
                                              core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = _bwa_version(self._execpath)
        return self._version

    def run(self, assets, parameters=tuple()):
        # samtools is a dependency. In the absence of features in RRT that would handle secondary depencies,
        # there is a simple assert here to prevent cryptic failures if the software is not on the system.
        samtools = 'samtools'
        assert environment.Executable.ispresent(self._execpath)
        assert environment.Executable.ispresent(samtools)
        
        source = assets.source
        indexfiles = tuple(name for cls, name in source.indexfilepattern.iterlistfiles())
        if len(indexfiles) == 0:
            raise ValueError("No BWA index files in %s" % source.indexfilepattern)

        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)
        if source.read2 is None:
            # single reads
            cmd_sub = ('mem', (source.indexfilepattern.name, source.read1.name))
        else:
            cmd_sub = ('mem', (source.indexfilepattern.name, source.read1.name, source.read2.name))

        # Gzip-compression not supported by BWA 
        if source.read1.iscompressed or (source.read2 is not None and source.read2.iscompressed):
            raise NotImplementedError("BWA does not support gzip-compressed FASTQ file. On-the-fly might eventually be implemented in the future.")
        # build command line
        cmd_align = ['%s' % self._execpath]
        cmd_align.append(cmd_sub[0])
        cmd_align.extend(parameters)
        cmd_align.extend(cmd_sub[1])
        
        cmd_sam2bam = (samtools, 'view', '-bS', '-')

        cmd = cmd_align + ['|', ] + list(cmd_sam2bam) + ['>',] + [assets.target.alignment.name, ] 
        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull, \
             open(assets.target.alignment.name, "w") as fh_out, \
             core.Popen(cmd_align, stdout=subprocess.PIPE, stderr=fnull) as p_align:
            # we are done yet: the output should be BAM, not SAM
            with core.Popen(cmd_sam2bam,
                            stdin=p_align.stdout,
                            stdout=fh_out,
                            stderr=fnull) as p_sam2bam:
                p_sam2bam.communicate()[0]
                returncode = p_sam2bam.returncode
        return (cmd, returncode)

        cmd.append(assets.target.alignment.name)
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
        # at this point, the output is in SAM while when want BAM
        return (cmd, returncode)


def _star_version(execpath):
    # How to get the version number courtesy of Brant Peterson.
    d = tempfile.mkdtemp()
    cmd = [execpath, '--outFileNamePrefix', d + os.path.sep]
    with open(os.devnull, 'w') as fnull:
        try:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd,
                                         stderr=fnull,
                                         stdout=fnull )
        except OSError as ose:
            raise UnifexError("""Command: %s
            %s""" % (' '.join(cmd), ose))
    # STAR is returning 102 (woohoo....)
    assert returncode==102, "STAR should have been returning 102 (and it did not)"
    # now dig the information out
    with open(os.path.join(d, "Log.out"), "r") as logfile:
        row = logfile.readline()
    shutil.rmtree(d)
    m = re.match('STAR svn revision compiled=(.+)', row)
    if m is None:
        m = re.match('STAR version=(.+)', row)
        if m is None:
            raise ValueError('The version number could be determined.')
    return m.groups()[0]


# --starindexer-begin
class StarIndex(IndexerAbstract):
    """
    STAR as an indexer.    
    """

    _name = 'star-index'
    _default_execpath = 'STAR'

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = _star_version(self._execpath)
        return self._version

    def run(self, assets, parameters=tuple()):
        """ The step on the assets given as parameters. """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' should be an instance of %s" % core.AssetsStep.__name__

        #if not os.path.exists(assets.target.indexfilepattern.name):
        os.mkdir(assets.target.indexfilepattern.name)
        # 
        basename_index = assets.target.indexfilepattern.name
        # command line is like:
        # STAR --runMode genomeGenerate \
        #     --genomeDir /path/to/GenomeDir \
        #     --genomeFastaFiles /path/to/genome/fasta1 /path/to/genome/fasta2        
        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        cmd.extend(['--runMode', 'genomeGenerate',
                    '--genomeDir',
                    assets.target.indexfilepattern.name,
                    '--genomeFastaFiles',
                    assets.source.reference.name])
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
        #FIXME: test return code ?
        return (cmd, returncode)
# --starindexer-end

class StarAlign(AlignerAbstract):
    """
    STAR as an aligner.    
    """

    _name = 'star-align'
    _default_execpath = 'STAR'

    class Assets(AssetsAligner):
        Target = core.assetfactory('Target', [core.AssetAttr('alignment', BAMFile, '')])

    def __init__(self, executable):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = _star_version(self._execpath)
        return self._version

    def run(self, assets, parameters=tuple()):
        # samtools is a dependency. In the absence of features in RRT that would handle secondary depencies,
        # there is a simple assert here to prevent cryptic failures if the software is not on the system.
        samtools = 'samtools'
        assert environment.Executable.ispresent(self._execpath)
        assert environment.Executable.ispresent(samtools)

        source = assets.source
        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        #--genomeDir /dir/STAR/Genome --readFilesIn /dir/*.fastq.gz --readFilesCommand zcat --outFileNamePrefix SampleA --runThreadN 20
        if source.read2 is None:
            # single reads
            cmd_sub = ['--readFilesIn', source.read1.name]
        else:
            cmd_sub = ['--readFilesIn', source.read1.name, source.read2.name]

        if source.read1.iscompressed:
            readfilescmd = 'zcat'
        else:
            readfilescmd = 'cat'

        cmd = [self._execpath]
        cmd.extend(parameters)
        cmd.extend(['--genomeDir', source.indexfilepattern.name, # reference index
                ])
        cmd.extend(cmd_sub)
        
        outdir = tempfile.mkdtemp(prefix='rrt_%s' % self._name,
                                  dir=os.path.dirname(assets.target.alignment.name))
        outfilename_prefix = os.path.join(outdir, 'star_output')
        cmd.extend([
            '--readFilesCommand', readfilescmd, #FIXME: check presence of zcat ?
            '--outFileNamePrefix', outfilename_prefix,
            '--outSAMunmapped', 'Within'
        ])
        with open(os.devnull, "w") as fnull:
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
            except TypeError as te:
                print(cmd)
                raise te
        if returncode == 0:
            # success, but we want a specific name for the alignment file
            starname = os.path.join(outfilename_prefix + 'Aligned.out.sam')

            # with open(os.devnull, "w") as fnull, \
            #      open(aligned_fn, "w") as fh_out:
            #     p_align = core.Popen(cmd_align, stdout=subprocess.PIPE, stderr=fnull)
            #     # we are done yet: the output should be BAM, not SAM
            #     p_sam2bam = core.Popen(cmd_sam2bam, stdin=p_align.stdout, stdout=fh_out, stderr=fnull)
            #     p_align.stdout.close()
            #     p_sam2bam.communicate()[0]
            #     returncode = p_sam2bam.returncode

            Assets = SamtoolsSamToBam.Assets
            assets = Assets(Assets.Source(SAMFile(starname)),
                            Assets.Target(BAMFile(assets.target.alignment.name)))
            s2b = SamtoolsSamToBam(samtools)
            s2b_cmd, s2b_returncode = s2b.run(assets)
                #
            #assets.target.bam
                #
                
        return (cmd, returncode)


class GsnapIndex(IndexerAbstract):
    """
    Gsnap as an indexer.
    
    Should the way it is run on the command line change, a child class
    will be implemented (and the relevant methods be overriden).
    """

    _name = 'gsnap-index'
    _default_execpath = 'gmap_build'

    def __init__(self, executable=None):
        """        
        :param execpath: path to the executable
        :type param: :class:`str`
        """
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath]
            with open(os.devnull, "w") as fnull:
                try:
                    logger.debug(subprocess.list2cmdline(cmd))
                    proc = core.Popen(cmd,
                                      stdout=subprocess.PIPE,
                                      stderr=fnull)
                except OSError as ose:
                    raise UnifexError("""Command: %s
                    %s""" % (' '.join(cmd), ose))
            m = None
            for row in proc.stdout:
                m = re.match(b'^.+? version ([0-9-]+)\.$', row)
                if m is not None:
                    break
            if m is None:
                raise Exception('The version number could not be extracted.')
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        """ Run the step on the assets given as parameters. 
        :param assets: the assets to run the step with.
        :param parameters: optional parameters as a sequence at the exception of the assets
        :type parameters: a sequence. """
        assert environment.Executable.ispresent(self._execpath)
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' must be an %s" % core.AssetsStep.__name__

        if not os.path.exists(assets.target.indexfilepattern.name):
            os.mkdir(assets.target.indexfilepattern.name)

        basename_index = assets.target.indexfilepattern.name
        #makefile = 'Makefile.reference'
        #makefile_path = os.path.join(basename_index, makefile)
        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        cmd.extend((
            '-d', 'reference',
            '-D', assets.target.indexfilepattern.name,
            assets.source.reference.name))
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.call(cmd, stdout = fnull, stderr = fnull)
        #FIXME: test return code ?
        return (cmd, returncode)


class GsnapAlign(AlignerAbstract):
    """
    Gsnap as an aligner.
    
    Should the way it is run on the command line change, a child class should
    be implemented (and the relevant methods be overriden).
    """

    _name = 'gsnap-align'
    _default_execpath = 'gsnap'

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            with open(os.devnull, "w") as fnull:
                cmd = [self._execpath]
                try:
                    logger.debug(subprocess.list2cmdline(cmd))
                    proc = core.Popen(cmd,
                                      stdout=fnull,
                                      stderr=subprocess.PIPE)
                except OSError as ose:
                    raise UnifexError("""Command: %s
                    %s""" % (' '.join(cmd), ose))
            row = next(proc.stderr)
            m = re.match(b'^.+ version ([^ ]+) .+$', row)
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        # samtools is a dependency. In the absence of features in RRT that would handle secondary depencies,
        # there is a simple assert here to prevent cryptic failures if the software is not on the system.
        samtools = 'samtools'
        assert environment.Executable.ispresent(self._execpath)
        assert environment.Executable.ispresent(samtools)

        # directory for the index
        assert isinstance(assets, core.AssetsStep), "The parameter 'assets' must be an %s" % core.AssetsStep.__name__

        basename_index = assets.source.indexfilepattern.name

        # different strings for the command line
        gunzip = False
        if assets.source.read2 is None:
            # single reads
            cmd_sub = (assets.source.read1.name, )
            if assets.source.read1.iscompressed:
                gunzip = True
        else:
            cmd_sub = (assets.source.read1.name, assets.source.read2.name)
            if assets.source.read1.iscompressed:
                if assets.source.read2.iscompressed:
                    gunzip = True
                else:
                    raise ValueError('With GSNAP FASTQ can either be all gzip-compressed or all gzip-uncompressed.')
            else:
                # gunzip is already set to False
                pass

        # build command line
        cmd_align = [self._execpath,]
        cmd_align.extend(parameters)
        cmd_align.extend((
            '-d', 'reference', #FIXME: move to parameters + way to have default parameters clearly exposed to higher-level layers
            '-A', 'sam',
            '-D', basename_index))
        if gunzip:
            cmd_align.append('--gunzip')
        cmd_align.extend(cmd_sub)

        cmd_sam2bam = (samtools, 'view', '-bS', '-')

        cmd = cmd_align + ['|', ] + list(cmd_sam2bam) + ['>',] + [assets.target.alignment.name, ] 
        logger.debug(subprocess.list2cmdline(cmd))
        with open(os.devnull, "w") as fnull, \
             open(assets.target.alignment.name, "w") as fh_out, \
             core.Popen(cmd_align, stdout=subprocess.PIPE, stderr=fnull) as p_align:
            # we are done yet: the output should be BAM, not SAM
            with core.Popen(cmd_sam2bam,
                            stdin=p_align.stdout,
                            stdout=fh_out,
                            stderr=fnull) as p_sam2bam:
                p_align.stdout.close()
                p_sam2bam.communicate()[0]
                returncode = p_sam2bam.returncode
        return (cmd, returncode)

class TopHat(core.StepAbstract):
    """
    TopHat
    th = TopHat(execpath)
    
    :param execpath: path to the executable

    """

    _name = 'tophat'
    _default_execpath = 'tophat'
    activities = (ACTIVITY.ALIGN, )

    class Assets(AssetsAligner):
        Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', SavedBowtieIndex, ''),
                                              core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])
        Target = core.assetfactory('Target', [core.AssetAttr('alignment', BAMFile, '')])

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = [self._execpath, '--version']
            try:
                logger.debug(subprocess.list2cmdline(cmd))
                res = subprocess.check_output(cmd)
            except OSError as ose:
                raise UnifexError("""Command: %s
                %s""" % (' '.join(cmd), ose))

            m = re.match(b'^.+? v([0-9][^ \n]+).*', res)
            if m is None:
                raise RuntimeError('Could not find the version number.')
            self._version = m.groups()[0]
        return self._version

    def run(self, assets, parameters=tuple()):
        assert environment.Executable.ispresent(self._execpath)
        source = assets.source
        indexfiles = tuple(name for cls, name in source.indexfilepattern.iterlistfiles())
        if len(indexfiles) == 0:
            raise ValueError("No bowtie1 index files in %s" % source.indexfilepattern)

        # FIXME: checks already done at the asset level. Code clean up to consider...
        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        if source.read2 is None:
            # single reads
            cmd_sub = (source.read1.name, )
        else:
            cmd_sub = (source.read1.name, source.read2.name)

        # build command line
        # notes:

        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        # TopHat saves results into a directory, and the alignment is called "accepted_hits.bam"
        output_dir = os.path.dirname(assets.target.alignment.name)
        if output_dir == '':
            output_dir = '.'
        cmd.extend(('-o', output_dir))
        cmd.append(source.indexfilepattern.name)
        cmd.extend(cmd_sub)
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd, stdout = fnull, stderr = fnull)        
        source_name = 'accepted_hits.bam'
        if source_name != assets.target.alignment.name:
            #FIXME: test for accidental overwrite ?
            #if os.path.exists(assets.target.alignment.name):
            shutil.move(os.path.join(output_dir, source_name), assets.target.alignment.name)
        return (cmd, returncode)

class TopHat2(TopHat):
    """
    TopHat2
    th = TopHat2(execpath)
    
    :param execpath: path to the executable

    """

    _name = 'tophat2'
    _default_execpath = 'tophat2'

    class Assets(TopHat.Assets):
        Source = core.assetfactory('Source', [core.AssetAttr('indexfilepattern', SavedBowtie2Index, ''),
                                              core.AssetAttr('read1', FASTQPossiblyGzipCompressed, ''),
                                              core.AssetAttr('read2', FASTQPossiblyGzipCompressed, '', allownone=True)])


    def run(self, assets, parameters=tuple()):
        assert environment.Executable.ispresent(self._execpath)
        source = assets.source
        #FIXME: TopHat2 can use bowtie or bowtie2
        indexfiles = tuple(name for cls, name in source.indexfilepattern.iterlistfiles())
        if len(indexfiles) == 0:
            raise ValueError("No bowtie2 index files in %s" % source.indexfilepattern)

        # FIXME: checks already done at the asset level. Code clean up to consider...
        # different strings for the command line
        if source.read1 is None or not isinstance(source.read1, core.SavedEntityAbstract):
            raise ValueError("Incorrect value %s for read1" % source.read1)

        if source.read2 is None:
            # single reads
            cmd_sub = (source.read1.name, )
        else:
            cmd_sub = ('%s,%s' % (source.read1.name, source.read2.name), )

        # build command line
        # notes:

        cmd = ['%s' % self._execpath]
        cmd.extend(parameters)
        # TopHat saves results into a directory, and the alignment is called "accepted_hits.bam"
        output_dir = os.path.dirname(assets.target.alignment.name)
        if output_dir == '':
            output_dir = '.'
        cmd.extend(('-o', output_dir))
        cmd.append(source.indexfilepattern.name)
        cmd.extend(cmd_sub)
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd, stdout = fnull, stderr = fnull)        
        source_name = 'accepted_hits.bam'
        if source_name != assets.target.alignment.name:
            #FIXME: test for accidental overwrite ?
            #if os.path.exists(assets.target.alignment.name):
            shutil.move(os.path.join(output_dir, source_name), assets.target.alignment.name)
        return (cmd, returncode)


class SamtoolsExtractUnaligned(core.StepAbstract):
    """
    Extract unaligned reads
    
    :param execpath: path to the executable

    """

    _name = 'samtools-extractunaligned'
    _default_execpath = 'samtools'


    class Assets(core.AssetsStep):
        """
        Assets for :class:`SamtoolsExtractUnaligned`
        
        """
        Source = core.assetfactory('Source', [core.AssetAttr('bamfile', BAMFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('unaligned', BAMFile, '')])

    activities = (ACTIVITY.UTILITY, )

    def __init__(self, executable=None):
        warnings.warn('SamtoolsExtracUnaliged is deprecated, use SamtoolsFilter', DeprecationWarning)
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = samtools_getversion(self._execpath)
        return self._version
        
    def run(self, assets, parameters = tuple()):
        assert environment.Executable.ispresent(self._execpath)
        cmd = [self._execpath, 'view', '-b', '-f', '4', 
               '-o', assets.target.unaligned.name, assets.source.bamfile.name, ]
        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        return (cmd, returncode)



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import unittest
from railroadtracks import core
from railroadtracks import unifex


class UnifexTestCase(unittest.TestCase):
    def test__extract_argdict(self):
        # invalid one arg
        args = ('/bar/baz.fq', )
        self.assertRaises(ValueError, unifex._extract_argdict, args)

        # one arg
        arglist = ('foo=/bar/baz.fq', )
        d = unifex._extract_argdict(arglist)
        self.assertEqual(set(('foo', )), set(d.keys()))

        # two args
        arglist = ('foo=/bar/baz.fq', 'bar=/baz/foo.bam')
        d = unifex._extract_argdict(arglist)
        self.assertEqual(set(('foo', 'bar')), set(d.keys()))

        # three args
        arglist = (('a=123', 'b=abc', 'a=456'))
        argdict = unifex._extract_argdict(arglist)
        # 
        self.assertEqual(set(argdict.keys()), set(('a','b')))
        self.assertEqual(argdict['a'], ['123','456'])
        self.assertEqual(argdict['b'], ['abc',])
        # test round trip
        arglist2 = unifex._extract_arglist(argdict)
        self.assertEqual(list(arglist).sort(), list(arglist2).sort())

    def test_build_AssetSet(self):
        values = (('abc',), ('123','456'))
        AssetSet = core.assetfactory('AssetSet', [core.AssetAttr('afile', core.File,''),
                                                  core.AssetAttr('asequenceoffiles', core.FileSequence, '')])
        assetset = unifex.build_AssetSet(AssetSet, values)
        #FIXME: where is the test ?
        #FIXME: how are allownone=True assets handled ?

        # values not allowed to be None cannot be None
        values = (None, ('123','456'))
        self.assertRaises(AssertionError, unifex.build_AssetSet, AssetSet, values)

        values = (('abc',), ('123','456'))
        AssetSet = core.assetfactory('AssetSet', [core.AssetAttr('afile', core.File,'', allownone=True),
                                                  core.AssetAttr('asequenceoffiles', core.FileSequence, '')])
        assetset = unifex.build_AssetSet(AssetSet, values)
        values = (None, ('123','456'))
        assetset = unifex.build_AssetSet(AssetSet, values)


    def test_AssetSet_repr(self):
        values = (('abc',), ('123','456'))
        AssetSet = core.assetfactory('AssetSet', [core.AssetAttr('afile', core.File,''),
                                                  core.AssetAttr('asequenceoffiles', core.FileSequence, '')])
        assetset = unifex.build_AssetSet(AssetSet, values)
        res = str(assetset)
        self.assertTrue(isinstance(res, str))



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import csv
import os
import re
import tempfile
import logging
import warnings
import shutil
import abc
import argparse
logger = logging.getLogger(__name__)
import subprocess
from railroadtracks import core, environment
from railroadtracks.core import SavedEntityAbstract, File, FileSequence

from six import with_metaclass

import sys
if sys.version_info[0] < 3:
    def astring(name):
        return isinstance(name, str) or isinstance(name, unicode)
else:
    from functools import reduce
    def astring(name):
        return isinstance(name, str) or isinstance(name, bytes)


class ACTIVITY(core.Enum):
    """ Activities that can be performed by the different steps modeled.
    (note: steps can combine several activities - the most obvious example is
    existing monolithic pipelines) """
    __order__ = 'CONVERTFORMAT'
    CONVERTFORMAT = 'Convert file formats'
    FILTERREADS = 'Filter reads'
    SORT = 'Sort'

class FASTAFile(File):
    """ FASTA file """
    _extension = ('.fa', '.fna', '.fasta', '.FASTA')

class GFFFile(File):
    """ GFF file """
    _extension = ('.gff', '.gtf')

class FASTQFile(File):
    """ FASTQ file """
    _extension = ('.fq', '.fastq')

class BAMFile(File):
    """ BAM file """
    _extension = ('.bam', )

class SAMFile(File):
    """ SAM file """
    _extension = ('.sam', )

class BEDFile(File):
    """ BED file """
    _extension = ('.bed', )

class FASTQPossiblyGzipCompressed(FASTQFile):
    _extension = tuple(y for ext in FASTQFile._extension for y in (ext, ext+'.gz'))
    @property
    def iscompressed(self):
        return self.name.endswith('.gz')

class SAMFileSortedByID(SAMFile):
    """ SAM file in which entries are sorted by read ID
    (this is required by tools such as HTSeq.). """
    pass

class BAMFileSortedByID(BAMFile):
    """ BAM file in which entries are sorted by read ID
    (this is required by tools such as HTSeq.). """
    pass

class CSVFile(File):
    """ CSV file """
    _extension = ('.csv', )
    def iter_entry(self):
        fh = open(self.name)
        csv_r = csv.reader(fh)
        for row in csv_r:
            yield row
        fh.close()

class TSVFile(File):
    """ Tab-separated files """
    _extension = ('.tsv', )
    def iter_entry(self):
        fh = open(self.name)
        csv_r = csv.reader(delimiter='\t')
        for row in csv_r:
            yield row
        fh.close()

# FIXME: should be Prefix, not Pattern
class FilePattern(SavedEntityAbstract):
    """
    Basename for a file.

    Some of the bioinformatics tools work with "basenames" and will produce or consume an arbitrary number
    of files.

    This object is able to represent a number of files having a given pattern.
    """

    def __init__(self, name):
        if name is not None:
            self._defined = True
            if not astring(name):
                if len(name) == 1:
                    name = next(iter(name))
                else:
                    raise ValueError('"name" should be a string.')
        self._name = name

    def _getname(self):
        assert self._defined, "The %s object is not defined." % type(self).__name__
        return self._name

    def _setname(self, name):
        assert not self._defined, "The %s object is already defined." % type(self).__name__
        self._name = name
        self._defined = True

    name = property(_getname, _setname, None,
                      '"Name" for a family of files. When not defined, this can be None and accessing it will raise an AssertionError (it can only be set then).')

    def __len__(self):
        return 1

    def __iter__(self):
        """ All files with ".name" as a prefix,
        and an extension matching the one in the list if it is defined, are returned """
        for x in self.iteritems():
            yield x[1]
        
    def iteritems(self):
        yield (type(self), self.name)

    def iterlistfiles(self):
        dn = os.path.dirname(self.name)
        if dn=='':
            dn = '.'
        bn = os.path.basename(self.name)
        for fn in os.listdir(dn):
            # if file extensions are defined, filter on extensions
            if (self._extension is not None) and (not any(fn.endswith(x) for x in self._extension)):
                continue            
            if fn.startswith(bn):
                yield (type(self), fn)
        

    def lastmodified(self):
        assert self._defined, "The %s object is not defined" % type(self).__name__
        it = iter(self)
        try:
            lasttime = next(it)
        except StopIteration:
            return None
        for fn in it:
            t = os.path.getmtime(os.path.join(self._dirname, fn))
            if t > lasttime:
                lasttime = t
        return lasttime


class CSVFileSequence(FileSequence):
    """ Sequence of CSV files """
    _type = CSVFile

class SAMFileSequence(FileSequence):
    """ Sequence of SAM files """
    _type = SAMFile


class SamtoolsFilter(core.StepAbstract):
    """
    Filter reads in a BAM file
    """
    _name = 'bam-filter'
    _default_execpath = 'samtools'

    class Assets(core.AssetsStep):
        """
        Assets for :class:`SamtoolsFilter`
        
        """
        Source = core.assetfactory('Source', [core.AssetAttr('bamfile', BAMFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('bamfile', BAMFile, '')])

    activities = (ACTIVITY.FILTERREADS, )

    parser = argparse.ArgumentParser(_name)
    FILTERVALUE_INFO = """
Flag        Chr     Description
0x0001      p       the read is paired in sequencing
0x0002      P       the read is mapped in a proper pair
0x0004      u       the query sequence itself is unmapped
0x0008      U       the mate is unmapped
0x0010      r       strand of the query (1 for reverse)
0x0020      R       strand of the mate
0x0040      1       the read is the first read in a pair
0x0080      2       the read is the second read in a pair
0x0100      s       the alignment is not primary
0x0200      f       the read fails platform/vendor quality checks
0x0400      d       the read is either a PCR or an optical duplicate
"""
    parser.add_argument('-f', '--filter-include',
                        dest = 'filter_include',
                        nargs = '+',
                        choices = ('0x0001','0x0002','0x0004','0x0008',
                                   '0x0010','0x0020','0x0040','0x0080',
                                   '0x0100','0x0200','0x0400','0x0800'),
                        help = """ Any of the following filter values can be specified.
Only return the reads satisfying all the filters.

                        %s""" % FILTERVALUE_INFO)
    parser.add_argument('-F', '--filter-exclude',
                        dest = 'filter_exclude',
                        nargs = '+',
                        choices = ('0x0001','0x0002','0x0004','0x0008',
                                   '0x0010','0x0020','0x0040','0x0080',
                                   '0x0100','0x0200','0x0400','0x0800'),
                        help = """ Any of the following filter values can be specified.
Only return the reads not satisfying any the filters.
                        %s""" % FILTERVALUE_INFO)


    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = samtools_getversion(self._execpath)
        return self._version
        
    def run(self, assets, parameters = tuple()):
        args = self.parser.parse_args(parameters)
        # FIXME: move checks out so can they can be made before calling 'run'
        if (args.filter_exclude is None) and (args.filter_include is None):
            raise ValueError('No filter specified.')

        cmd = [self._execpath, 'view', '-o', assets.target.bamfile.name]
        if (args.filter_exclude is not None):
            filter_exclude = reduce(lambda x,y: x^y, (x for x in args.filter_exclude))
            cmd.extend(('-F', filter_exclude))
        if (args.filter_include is not None):
            filter_include = reduce(lambda x,y: x^y, (x for x in args.filter_include))
            cmd.extend(('-f', filter_include))
            
        cmd.extend(('-b', assets.source.bamfile.name))

        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        if not os.path.exists(assets.source.bamfile.name):
            # target is missing. Suspected infamous 'bam.bam' issue
            bam_bam = assets.source.bamfile.name + '.bam'
            if os.path.exists(bam_bam):
                # '.bam.bam' issue
                warnings.warn("'.bam.bam' issue detected. Moving the product to intended target %s." % assets.source.samfile.name)
                shutil.move(bam_bam, assets.source.bamfile.name)
            else:
                raise Exception('The target %s is mysteriously missing.' % assets.source.samfile.name)
        return (cmd, returncode)


class SamtoolsSamToBam(core.StepAbstract):
    """
    Convert a SAM file into a BAM file
    """
    _name = 'sam-to-bam'
    _default_execpath = 'samtools'

    class Assets(core.AssetsStep):
        """
        Assets for :class:`SamtoolsSamToBam`
        
        """
        Source = core.assetfactory('Source', [core.AssetAttr('samfile', SAMFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('bamfile', BAMFile, '')])

    activities = (ACTIVITY.CONVERTFORMAT, )

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = samtools_getversion(self._execpath)
        return self._version
        
    def run(self, assets, parameters = tuple()):
        cmd = [self._execpath, 'view', '-o', assets.target.bamfile.name, '-Sb', assets.source.samfile.name, ]
        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        if not os.path.exists(assets.source.samfile.name):
            # target is missing. Suspected infamous 'bam.bam' issue
            bam_bam = assets.source.samfile.name + '.bam'
            if os.path.exists(bam_bam):
                # '.bam.bam' issue
                warnings.warn("'.bam.bam' issue detected. Moving the product to intended target %s." % assets.source.samfile.name)
                shutil.move(bam_bam, assets.source.samfile.name)
            else:
                raise Exception('The target %s is mysteriously missing.' % assets.source.samfile.name)
        return (cmd, returncode)


class SamtoolsBamToSam(core.StepAbstract):
    """
    Convert a BAM file into a SAM file
    """
    _name = 'bam-to-sam'
    _default_execpath = 'samtools'

    class Assets(core.AssetsStep):
        """
        Assets for :class:`SamtoolsBamToSam`
        
        """
        Source = core.assetfactory('Source', [core.AssetAttr('bamfile', BAMFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('samfile', SAMFile, '')])

    activities = (ACTIVITY.CONVERTFORMAT, )

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = samtools_getversion(self._execpath)
        return self._version
        
    def run(self, assets, parameters = tuple()):
        cmd = [self._execpath, 'view', '-h', '-o', assets.target.samfile.name, assets.source.bamfile.name, ]
        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        return (cmd, returncode)


def samtools_getversion(execpath):
    """ Return the version of 'samtools'. """

    cmd = [execpath,]
    logging.debug(cmd)
    m = None
    with core.Popen(cmd,
                    stderr=subprocess.PIPE) as proc:

        for row in proc.stderr:
            m = re.match(b'^Version: ([^ \n]+).*$', row)
            if m is not None:
                break
    if m is None:
        raise RuntimeError('Could not find the version number.')
    version = m.groups()[0]
    return version


def ensure_bam(filename):
    """ Ensure that a file is in the BAM format.
    If 'filename' is already a BAM file, return a filehandle, otherwise
    create a tempfile and return a filehandle to it.
    """
    # cheap test based on file extension
    if filename.endswith('.bam'):
        return open(filename, 'rb')
    elif filename.endswith('.sam'):
        samtools = environment.Executable('samtools')
        newfile = tempfile.NamedTemporaryFile(suffix='.bam')
        cmd = [samtools.path, 'view', '-o', newfile.name, '-Sb', filename, ]
        with open(os.devnull, 'w') as fnull: 
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        return newfile
    else:
        raise ValueError('The filename "%s" is not .sam neither .bam' % filename)

def ensure_sam(filename):
    """ Ensure that a file is in the SAM format.
    If 'filename' is already a SAM file, return a filehandle, otherwise
    create a tempfile and return a filehandle to it.
    """
    # cheap test based on file extension
    if filename.endswith('.bam'):
        samtools = environment.Executable('samtools')
        newfile = tempfile.NamedTemporaryFile(suffix='.sam')
        cmd = [samtools.path, 'view', '-o', newfile.name, '-h', filename, ]
        with open(os.devnull, 'w') as fnull: 
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd,
                                               stderr = fnull,
                                               stdout = fnull)
        return newfile
    elif filename.endswith('.sam'):
        return open(filename, 'r')
    else:
        raise ValueError('The filename "%s" is not .sam neither .bam' % filename)


class AssetsSorter(core.AssetsStep):
    Source = core.assetfactory('Source', [core.AssetAttr('alignedreads',BAMFile, '')])
    Target = core.assetfactory('Target', [core.AssetAttr('sortedbam', BAMFile, '')])

class SorterAbstract(with_metaclass(abc.ABCMeta, core.StepAbstract)):
    """

    A sorting step.

    """
    activities = (ACTIVITY.SORT, )
    Assets = AssetsSorter

class SamtoolsSorterByID(SorterAbstract):
    """
    """

    _name = 'samtools-sortbyid'
    _default_execpath = 'samtools'

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            self._version = samtools_getversion(self._execpath)
        return self._version

    def run(self, assets, parameters = tuple()):        
        # build command line
        cmd = ['%s' % self._execpath]
        cmd.append('sort')
        cmd.append('-n')
        cmd.append('-f') # otherwise 'samtools' is interpreting the output as a prefix rather than a complete file name
        cmd.extend(parameters)
        cmd.append('%s' % assets.source.alignedreads.name)
        cmd.append('%s' % assets.target.sortedbam.name) 
        with open(os.devnull, "w") as fnull:
            logger.debug(subprocess.list2cmdline(cmd))
            returncode = subprocess.check_call(cmd, 
                                               stdout = fnull,
                                               stderr = fnull)
        return (cmd, returncode)



class BedtoolsBamToFastq(core.StepAbstract):
    """
    Convert a BAM file into a FASTQ file
    """
    _name = 'BAM-to-FASTQ'
    _default_execpath = 'bedtools'

    class Assets(core.AssetsStep):
        """
        Assets for :class:`BedtoolsBamToFastq`
        
        """
        Source = core.assetfactory('Source', [core.AssetAttr('bamfile', BAMFile, '')])
        Target = core.assetfactory('Target', [core.AssetAttr('fastqfile', FASTQFile, '')])

    activities = (ACTIVITY.CONVERTFORMAT, )

    def __init__(self, executable=None):
        if executable is None:
            executable = type(self)._default_execpath
        self._execpath = executable
        self._version = None

    @property
    def version(self):
        if self._version is None:
            cmd = ('bedtools', '--version')
            output = subprocess.check_output(cmd)
            m = re.match("^bedtools v(.+)", output)
            self._version = m.groups()[0]
        return self._version
        
    def run(self, assets, parameters = tuple()):
        cmd = [self._execpath, 'bamtofastq',
               '-i', assets.source.bamfile.name, 
               '-fq', assets.target.fastqfile.name, ]
        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        return (cmd, returncode)


class BedtoolsBamToFastqPE(BedtoolsBamToFastq):


    _name = 'BAM-to-FASTQ-PE'

    class Assets(BedtoolsBamToFastq.Assets):
        Target = core.assetfactory('Target', [core.AssetAttr('fastqfile', FASTQFile, ''),
                                              core.AssetAttr('fastqfile2', FASTQFile, '')])

    activities = (ACTIVITY.CONVERTFORMAT, )
        
    def run(self, assets, parameters = tuple()):
        cmd = [self._execpath, 'bamtofastq',
               '-i', assets.source.bamfile.name, 
               '-fq', assets.target.fastqfile.name,
               '-fq2', assets.target.fastqfile2.name ]
        with open(os.devnull, 'w') as fnull: 
            logging.debug(cmd)
            returncode = subprocess.check_call(cmd,
                                               stdout = fnull,
                                               stderr = fnull)
        return (cmd, returncode)




#!/usr/bin/env python

# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import sys
if (sys.version_info[0] == 2) and (sys.version_info[1] < 7):
    print("Python>=2.7 is required.")
    sys.exit(1)

try:
    from setuptools import setup
except ImportError:
    print("The Python package 'setuptools' is required in order to install this package.")
    sys.exit(1)

from src import __version__

packagename='railroadtracks'
setup(name=packagename,
      version=__version__,
      description='Framework to connect computional steps with an emphasis on RNA-Seq',
      author='Laurent Gautier',
      author_email='laurent.gautier@novartis.com',
      license='License :: OSI Approved :: Apache Software License',
      url='https://github.com/Novartis/railroadtracks',
      package_dir={packagename: 'src'},
      package_data={packagename: ['src/cache.sql',
                                  'src/model/featurecount.R',
                                  'src/model/edge.R',
                                  'src/model/deseq.R',
                                  'src/model/deseq2.R',
                                  'src/model/limmavoom.R',
                                  'src/EF204940.fa',
                                  'src/ef204940.gff',
                                  'src/ef204940.gtf',
                                  'src/templates/*.html']},
      include_package_data=True,
      packages=[packagename, 
                packagename + '.easy',
                packagename + '.model',
                packagename + '.test'],
      install_requires=['flufl.enum', 'networkx', 'jinja2', 'six', 'enum34'] #FIXME: enum34 not needed if Python >= 3.4
     )



# Copyright 2014-2015 Novartis Institutes for Biomedical Research

# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at

#     http://www.apache.org/licenses/LICENSE-2.0

# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import tempfile, shutil, os, subprocess
from collections import namedtuple
import unittest

import railroadtracks
from railroadtracks import environment, rnaseq, easy

import railroadtracks.model.simulate
PHAGEFASTA = railroadtracks.model.simulate.PHAGEFASTA

class EasyTestCase(unittest.TestCase):

    def setUp(self):
        wd = tempfile.mkdtemp()
        self.wd = wd

    def tearDown(self):
        shutil.rmtree(self.wd)

    def test_FrozenNameSpace(self):
        kv = (('a', 1), ('b', 2))
        fz = easy.FrozenNamespace(kv)
        fz_keys = list(fz._fields)
        fz_keys.sort()
        for x,y in zip(kv, fz_keys):
            self.assertEqual(x[0],y)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')            
    def test_command_line(self):
        packagedir = os.path.dirname(railroadtracks.__file__)
        PHAGEFASTA = os.path.join(packagedir, 'EF204940.FASTA')
        bowtie2index_exec = environment.Executable('bowtie2-build')
        project = easy.Project(rnaseq, self.wd)
        bowtie2 = rnaseq.Bowtie2Build()
        assets = bowtie2.Assets(bowtie2.Assets.Source(rnaseq.FASTAFile(PHAGEFASTA)),
                                bowtie2.Assets.Target(rnaseq.SavedBowtie2Index(os.path.join(self.wd, 'foo'))))
        parameters = tuple()
        step_concrete_id = project.cache.add(bowtie2, assets, parameters = parameters)
        cmd = easy.tasks.command_line(project, step_concrete_id, bowtie2, assets, parameters = parameters)
        returncode = subprocess.check_call(cmd)
        self.assertEqual(0, returncode)

    def _test_EasyProject(self, cached):
        packagedir = os.path.dirname(railroadtracks.__file__)
        PHAGEFASTA = os.path.join(packagedir, 'EF204940.FASTA')
        bowtie2index_exec = environment.Executable('bowtie2-build')
        project = easy.Project(rnaseq, self.wd, cached=cached)
        # Initial number of steps is zero
        self.assertEqual(0, project.persistent_graph.nconcrete_steps)
        bowtie2index = rnaseq.Bowtie2Build(bowtie2index_exec.path)
        reference_fn = PHAGEFASTA
        task_index = project.add_task(bowtie2index, 
                                      bowtie2index.Assets(bowtie2index.Assets.Source(rnaseq.FASTAFile(reference_fn))))
        # Number of steps is one
        self.assertEqual(1, project.persistent_graph.nconcrete_steps)
        task_index_same = project.add_task(bowtie2index, 
                                           bowtie2index.Assets(bowtie2index.Assets.Source(rnaseq.FASTAFile(reference_fn))))
        # Number of steps should remain one (same assets and paremeters)
        self.assertEqual(1, project.persistent_graph.nconcrete_steps)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_EasyProject_NoCache(self):
        self._test_EasyProject(False)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_EasyProject_Cache(self):
        self._test_EasyProject(True)

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_Project_get_tasksoftype(self):
        packagedir = os.path.dirname(railroadtracks.__file__)
        PHAGEFASTA = os.path.join(packagedir, 'EF204940.FASTA')
        bowtie2index_exec = environment.Executable('bowtie2-build')
        project = easy.Project(rnaseq, self.wd)
        bowtie2index = rnaseq.Bowtie2Build(bowtie2index_exec.path)
        reference_fn = PHAGEFASTA

        for obj in (rnaseq.Bowtie2Build, bowtie2index):
            taskset = project.get_tasksoftype(obj)
            self.assertEqual(0, len(taskset))
        
        task_index = project.add_task(bowtie2index, 
                                      bowtie2index.Assets(bowtie2index.Assets.Source(rnaseq.FASTAFile(reference_fn))))
        
        for obj in (rnaseq.Bowtie2Build, bowtie2index, task_index):
            taskset = project.get_tasksoftype(obj)
            self.assertEqual(1, len(taskset))



    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_Task(self):
        project = easy.Project(rnaseq, self.wd)
        #
        packagedir = os.path.dirname(railroadtracks.__file__)
        PHAGEFASTA = os.path.join(packagedir, 'EF204940.FASTA')
        bowtie2index_exec = environment.Executable('bowtie2-build')
        bowtie2align_exec = environment.Executable('bowtie2')

        bowtie2index = rnaseq.Bowtie2Build(bowtie2index_exec.path)
        reference_fn = PHAGEFASTA
        task_index = project.add_task(bowtie2index, 
                                      bowtie2index.Assets(bowtie2index.Assets.Source(rnaseq.FASTAFile(reference_fn))))

        bowtie2align = rnaseq.Bowtie2(bowtie2align_exec.path)
        Assets = bowtie2align.Assets
        alignment_tasks = list()
        for read1_name, read2_name in (('f1_1.fq', 'f1_2.fq'),
                                       ('f2_1.fq', 'f2_2.fq')):
            assets = Assets(Assets.Source(task_index.call.assets.target.indexfilepattern, 
                                          rnaseq.FASTQPossiblyGzipCompressed(read1_name),
                                          rnaseq.FASTQPossiblyGzipCompressed(read2_name)),
                            None)
            alignment_tasks.append(project.add_task(bowtie2align, assets))

        # find all child
        res = task_index.all_child_tasks()
        s1 = set(x.task_id for x in alignment_tasks)
        s2 = set(x.task_id for x in res)
        self.assertEqual(0, len(s1.difference(s2)))
        self.assertEqual(0, len(s2.difference(s1)))

        # find the parents of an alignment
        align_task = next(iter(alignment_tasks))
        parent_taskset = align_task.parent_tasks()
        # only one parent (the indexing step)
        self.assertEqual(1, len(parent_taskset))
        parent_task = next(iter(parent_taskset))
        self.assertEqual(bowtie2index.activities, parent_task.call.step.activities)
        self.assertEqual(bowtie2index._name, parent_task.call.step._name)

        # split by status
        # (first create a TaskSet)
        alignment_taskset = easy.TaskSet()
        for t in alignment_tasks:
            alignment_taskset.add(t)
        taskset_groups = alignment_taskset.split_by_status()
        self.assertEqual(1, len(taskset_groups)) # only one status here

        # root tasks
        root_tasks = align_task.primordial_tasks()
        # check that the indexing step is indeed one them
        self.assertTrue(task_index.task_id in (x.task_id for x in root_tasks))

    @unittest.skipIf(not environment.Executable.ispresent('bowtie2-build'),
                     'bowtie2-build is not in the PATH')
    def test_Task_withNone(self):
        project = easy.Project(rnaseq, self.wd)
        #
        packagedir = os.path.dirname(railroadtracks.__file__)
        PHAGEFASTA = os.path.join(packagedir, 'EF204940.FASTA')
        bowtie2index_exec = environment.Executable('bowtie2-build')
        bowtie2align_exec = environment.Executable('bowtie2')

        bowtie2index = rnaseq.Bowtie2Build(bowtie2index_exec.path)
        reference_fn = PHAGEFASTA
        task_index = project.add_task(bowtie2index, 
                                      bowtie2index.Assets(bowtie2index.Assets.Source(rnaseq.FASTAFile(reference_fn))))

        bowtie2align = rnaseq.Bowtie2(bowtie2align_exec.path)
        Assets = bowtie2align.Assets
        alignment_tasks = list()
        for read1_name, read2_name in (('f1_1.fq', None),
                                       ('f2_1.fq', None)):
            assets = Assets(Assets.Source(task_index.call.assets.target.indexfilepattern, 
                                          rnaseq.FASTQPossiblyGzipCompressed(read1_name),
                                          read2_name),
                            None)
            alignment_tasks.append(project.add_task(bowtie2align, assets))

        # find all child
        res = task_index.all_child_tasks()
        s1 = set(x.task_id for x in alignment_tasks)
        s2 = set(x.task_id for x in res)
        self.assertEqual(0, len(s1.difference(s2)))
        self.assertEqual(0, len(s2.difference(s1)))

        # find the parents of an alignment
        task = next(iter(alignment_tasks))
        parent_taskset = task.parent_tasks()
        # only one parent (the indexing step)
        self.assertEqual(1, len(parent_taskset))
        parent_task = next(iter(parent_taskset))
        self.assertEqual(bowtie2index.activities, parent_task.call.step.activities)
        self.assertEqual(bowtie2index._name, parent_task.call.step._name)



class IterativeExecutionTestCase(unittest.TestCase):
    def setUp(self):
        wd = tempfile.mkdtemp()
        self.wd2 = tempfile.mkdtemp()
        self.project = easy.Project(rnaseq, wd)
        bowtie2index = rnaseq.Bowtie2Build()
        reference_fn = PHAGEFASTA
        Assets = bowtie2index.Assets
        task = self.project.add_task(bowtie2index, 
                                     Assets(Assets.Source(rnaseq.FASTAFile(reference_fn))))
        self.task = task
        self.bowtie2index = bowtie2index

    def tearDown(self):
        shutil.rmtree(self.project.wd)
        shutil.rmtree(self.wd2)

    def _create_fastq(self, n, d):
        reference_fn = PHAGEFASTA
        with open(reference_fn) as fasta_fh:
            reference = next(railroadtracks.model.simulate.readfasta_iter(fasta_fh))

        ReadPair = namedtuple('ReadPair', 'read1 read2')
        pairnames = tuple(ReadPair(os.path.join(d, 'A_%i_1.fq' % p_i), 
                                   os.path.join(d, 'A_%i_2.fq' % p_i)) for p_i in range(1, n+1))
        for read1_fn, read2_fn in pairnames:
            with open(read1_fn, 'wb') as read1_fh, \
                 open(read2_fn, 'wb') as read2_fh:
                railroadtracks.model.simulate.randomPEreads(read1_fh, read2_fh, reference)
        return pairnames

    def testMap(self):

        mapper = easy.IterativeExecution()

        ts = easy.TaskSet()
        ts.add(self.task)

        mapper.map(ts)

        n_fastq = 3
        pairnames = self._create_fastq(n_fastq, self.wd2)

        index_task = self.task
        bowtie2 = rnaseq.Bowtie2()
        Assets = bowtie2.Assets
        FASTQ = rnaseq.FASTQPossiblyGzipCompressed
        ts = easy.TaskSet()

        for read1_fn, read2_fn in pairnames:
            task = self.project.add_task(bowtie2,
                                         Assets(Assets.Source(index_task.call.assets.target.indexfilepattern,
                                                              FASTQ(read1_fn),
                                                              FASTQ(read2_fn))))
            ts.add(task)
        self.assertEqual(3, ts.status()[railroadtracks.easy._TASK_TODO])
        mapper.map(ts)
        self.assertEqual(3, ts.status()[railroadtracks.easy._TASK_DONE])


    def testIter_chunk(self):
        ts = easy.TaskSet()
        ts.add(self.task)
        Assets = self.bowtie2index.Assets
        reference_fn = PHAGEFASTA
        for tag in range(2, 11):
            task = self.project.add_task(self.bowtie2index, 
                                         Assets(Assets.Source(rnaseq.FASTAFile(reference_fn))),
                                         tag=tag)
            ts.add(task)

        self.assertEqual(10, len(ts))
        chunks = tuple(ts.iter_chunks(3))
        for i in range(3):
            self.assertEqual(3, len(chunks[i]))
        self.assertEqual(1, len(chunks[3]))
